"""Cryoscope experiment."""

from dataclasses import dataclass, field

import cma
import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import scipy
import scipy.signal
from plotly.subplots import make_subplots
from qibolab import (
    AcquisitionType,
    AveragingMode,
    Delay,
    Platform,
    Pulse,
    PulseSequence,
    Rectangular,
)
from scipy.optimize import curve_fit
from scipy.signal import lfilter

from qibocal import update
from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.config import log
from qibocal.protocols.ramsey.utils import fitting
from qibocal.protocols.utils import table_dict, table_html

# TODO: remove hard-coded QM parameters
FEEDFORWARD_MAX = 2 - 2**-16
"""Maximum feedforward tap value"""
FEEDBACK_MAX = 1 - 2**-20
"""Maximum feedback tap value"""
SAMPLING_RATE = 1
"""Instrument sampling rate in GSamples"""


@dataclass
class CryoscopeParameters(Parameters):
    """Cryoscope runcard inputs."""

    duration_min: int
    """Minimum flux pulse duration."""
    duration_max: int
    """Maximum flux duration start."""
    duration_step: int
    """Flux pulse duration step."""
    flux_pulse_amplitude: float
    """Flux pulse amplitude."""
    fir: int = 20
    """Number of feedforward taps to be optimized after IIR."""
    unrolling: bool = True


@dataclass
class CryoscopeResults(Results):
    """Cryoscope outputs."""

    fitted_parameters: dict[tuple[QubitId, str], list[float]] = field(
        default_factory=dict
    )
    """Fitted <X> and <Y> for each qubit."""
    detuning: dict[QubitId, list[float]] = field(default_factory=dict)
    """Expected detuning."""
    amplitude: dict[QubitId, list[float]] = field(default_factory=dict)
    """Flux amplitude computed from detuning."""
    step_response: dict[QubitId, list[float]] = field(default_factory=dict)
    """Waveform normalized to 1."""
    exp_amplitude: dict[QubitId, list[float]] = field(default_factory=dict)
    """A parameters for the exp decay approximation"""
    tau: dict[QubitId, list[float]] = field(default_factory=dict)
    """time decay constant in exp decay approximation"""
    feedforward_taps: dict[QubitId, list[float]] = field(default_factory=dict)
    """feedforward taps"""
    feedforward_taps_iir: dict[QubitId, list[float]] = field(default_factory=dict)
    """feedforward taps for IIR"""
    feedback_taps: dict[QubitId, list[float]] = field(default_factory=dict)
    """feedback taps"""

    def __contains__(self, key):
        return key in self.detuning


CryoscopeType = np.dtype([("duration", int), ("prob_1", np.float64)])
"""Custom dtype for Cryoscope."""


def generate_sequences(
    platform: Platform,
    qubit: QubitId,
    duration: int,
    params: CryoscopeParameters,
) -> tuple[PulseSequence, PulseSequence]:
    """Compute sequences at fixed duration of flux pulse for <X> and <Y>"""
    native = platform.natives.single_qubit[qubit]

    drive_channel, ry90 = native.R(theta=np.pi / 2, phi=np.pi / 2)[0]
    _, rx90 = native.R(theta=np.pi / 2)[0]
    ro_channel, ro_pulse = native.MZ()[0]
    flux_channel = platform.qubits[qubit].flux

    flux_pulse = Pulse(
        duration=duration,
        amplitude=params.flux_pulse_amplitude,
        envelope=Rectangular(),
    )

    # create the sequences
    sequence_x, sequence_y = PulseSequence(), PulseSequence()

    sequence_x.extend(
        [
            (drive_channel, ry90),
            (flux_channel, Delay(duration=ry90.duration)),
            (flux_channel, flux_pulse),
            (drive_channel, Delay(duration=params.duration_max + 100)),
            (drive_channel, ry90),
            (
                ro_channel,
                Delay(
                    duration=ry90.duration + params.duration_max + 100 + ry90.duration
                ),
            ),
            (ro_channel, ro_pulse),
        ]
    )

    sequence_y.extend(
        [
            (drive_channel, ry90),
            (flux_channel, Delay(duration=rx90.duration)),
            (flux_channel, flux_pulse),
            (drive_channel, Delay(duration=params.duration_max + 100)),
            (drive_channel, rx90),
            (
                ro_channel,
                Delay(
                    duration=ry90.duration + params.duration_max + 100 + rx90.duration
                ),
            ),
            (ro_channel, ro_pulse),
        ]
    )
    return sequence_x, sequence_y


@dataclass
class CryoscopeData(Data):
    """Cryoscope acquisition outputs."""

    flux_pulse_amplitude: float
    """Flux pulse amplitude."""
    fir: int
    """Number of feedforward taps to be optimized after IIR."""
    flux_coefficients: dict[QubitId, list[float]] = field(default_factory=dict)
    """Flux - amplitude relation coefficients obtained from flux_amplitude_frequency routine"""
    filters: dict[QubitId, float] = field(default_factory=dict)
    """Check if there are filters already."""
    data: dict[tuple[QubitId, str], npt.NDArray[CryoscopeType]] = field(
        default_factory=dict
    )

    def has_filters(self, qubit: QubitId) -> bool:
        """Checking if for qubit there are already filters."""
        try:
            return len(self.filters[qubit]) > 0
        except AttributeError:
            return False


def _acquisition(
    params: CryoscopeParameters,
    platform: Platform,
    targets: list[QubitId],
) -> CryoscopeData:
    """Acquisition for cryoscope experiment.

    The following sequence is played for each qubit.

    drive    --- RY90 ------------------- RY90 -------
    flux     --------- FluxPulse(t) ------------------
    readout  ----------------------------------- MZ --

    The previous sequence measures <X>, to measure <Y> the second drive pulse
    is replaced with RX90.
    The delay between the two pi/2 pulses is fixed at t_max (maximum duration of flux pulse)
    + 100 ns (following the paper).
    """
    data = CryoscopeData(
        fir=params.fir,
        flux_pulse_amplitude=params.flux_pulse_amplitude,
    )

    for qubit in targets:
        assert (
            platform.calibration.single_qubits[qubit].qubit.flux_coefficients
            is not None
        ), (
            f"Cannot run cryoscope without flux coefficients, run cryoscope amplitude on qubit {qubit} before the cryoscope"
        )

        data.flux_coefficients[qubit] = platform.calibration.single_qubits[
            qubit
        ].qubit.flux_coefficients
        data.filters[qubit] = platform.config(platform.qubits[qubit].flux).filter

    sequences_x = []
    sequences_y = []

    duration_range = np.arange(
        params.duration_min, params.duration_max, params.duration_step
    )

    for duration in duration_range:
        sequence_x = PulseSequence()
        sequence_y = PulseSequence()

        for qubit in targets:
            qubit_sequence_x, qubit_sequence_y = generate_sequences(
                platform, qubit, duration, params
            )
            sequence_x += qubit_sequence_x
            sequence_y += qubit_sequence_y

        sequences_x.append(sequence_x)
        sequences_y.append(sequence_y)

    options = dict(
        nshots=params.nshots,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.CYCLIC,
    )

    if params.unrolling:
        results_x = platform.execute(sequences_x, **options)
        results_y = platform.execute(sequences_y, **options)
    else:
        results_x = [
            platform.execute([sequence], **options) for sequence in sequences_x
        ]
        results_y = [
            platform.execute([sequence], **options) for sequence in sequences_y
        ]

    for measure, results, sequence in zip(
        ["MX", "MY"], [results_x, results_y], [sequences_x, sequences_y]
    ):
        for i, (duration, sequence) in enumerate(zip(duration_range, sequence)):
            for qubit in targets:
                ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[
                    -1
                ]
                result = (
                    results[ro_pulse.id]
                    if params.unrolling
                    else results[i][ro_pulse.id]
                )
                data.register_qubit(
                    CryoscopeType,
                    (qubit, measure),
                    dict(
                        duration=np.array([duration]),
                        prob_1=result,
                    ),
                )

    return data


def exponential_params(step_response, acquisition_time):
    t = np.arange(0, acquisition_time, 1)
    init_guess = [1, 10, 1]
    target = np.ones(len(t))

    def expmodel(t, tau, exp_amplitude, g):
        return step_response / (g * (1 + exp_amplitude * np.exp(-t / tau)))

    popt, _ = curve_fit(expmodel, t, target, p0=init_guess)

    return popt


def filter_calc(params):
    tau, exp_amplitude, _ = params
    alpha = 1 - np.exp(-1 / (SAMPLING_RATE * tau * (1 + exp_amplitude)))
    k = (
        exp_amplitude / ((1 + exp_amplitude) * (1 - alpha))
        if exp_amplitude < 0
        else exp_amplitude / (1 + exp_amplitude - alpha)
    )
    b0 = 1 - k + k * alpha
    b1 = -(1 - k) * (1 - alpha)
    a0 = 1
    a1 = -(1 - alpha)

    feedback_taps = np.array([a0, a1])
    feedforward_taps = np.array([b0, b1])

    if np.any(np.abs(feedback_taps) > FEEDBACK_MAX):
        feedback_taps[feedback_taps > FEEDBACK_MAX] = FEEDBACK_MAX
        feedback_taps[feedback_taps < -FEEDBACK_MAX] = -FEEDBACK_MAX

    return feedback_taps.tolist(), feedforward_taps.tolist()


def _fit(data: CryoscopeData) -> CryoscopeResults:
    """Postprocessing for cryoscope experiment.

    From <X> and <Y> we compute the expecting step response.
    The complex data <X> + i <Y> are demodulated using the frequency found
    by fitting a sinusoid to both <X> and <Y>.
    Next, the phase is computed and finally the detuning using a savgol_filter.
    The "real" detuning is computed by reintroducing the demodulation frequency.
    Finally, using the parameters given by the flux_amplitude_frequency experiment,
    we compute the expected flux_amplitude by inverting the formula:

    f = c_1 A^2 + c_2 A + c_3

    where f is the detuning and A is the flux amplitude.
    The step response is computed by normalizing the amplitude by its value computed above.
    For some of the manipulations see: https://github.com/DiCarloLab-Delft/PycQED_py3/blob/c4279cbebd97748dc47127e56f6225021f169257/pycqed/analysis/tools/cryoscope_tools.py#L73
    """

    nyquist_order = 0

    fitted_parameters = {}
    detuning = {}
    amplitude = {}
    step_response = {}
    alpha = {}
    g = {}
    time_decay = {}
    feedforward_taps_iir = {}
    feedforward_taps = {}
    feedback_taps = {}
    for qubit, setup in data.data:
        qubit_data = data[qubit, setup]
        x = qubit_data.duration
        y = 1 - 2 * qubit_data.prob_1

        popt, _ = fitting(x, y)

        fitted_parameters[qubit, setup] = popt

    qubits = np.unique([i[0] for i in data.data]).tolist()

    for qubit in qubits:
        sampling_rate = 1 / (x[1] - x[0])
        X_exp = 1 - 2 * data[(qubit, "MX")].prob_1
        Y_exp = 1 - 2 * data[(qubit, "MY")].prob_1

        norm_data = X_exp + 1j * Y_exp

        # demodulation frequency found by fitting sinusoidal
        demod_freq = -fitted_parameters[qubit, "MY"][2] / 2 / np.pi * sampling_rate

        # to be used in savgol_filter
        derivative_window_length = 7 / sampling_rate
        derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
        derivative_window_size += (derivative_window_size + 1) % 2

        # find demodulatation frequency
        demod_data = np.exp(2 * np.pi * 1j * x * demod_freq) * (norm_data)

        # compute phase
        phase = np.unwrap(np.angle(demod_data))

        # compute detuning
        raw_detuning = (
            scipy.signal.savgol_filter(
                phase / (2 * np.pi),
                window_length=derivative_window_size,
                polyorder=2,
                deriv=1,
            )
            * sampling_rate
        )

        # real detuning (reintroducing demod_freq)
        detuning[qubit] = (
            raw_detuning - demod_freq + sampling_rate * nyquist_order
        ).tolist()

        # params from flux_amplitude_frequency_protocol
        params = data.flux_coefficients[qubit]

        # invert frequency amplitude formula
        p = np.poly1d(params)
        amplitude[qubit] = [max((p - freq).roots).real for freq in detuning[qubit]]

        # compute step response
        step_response[qubit] = (
            np.array(amplitude[qubit]) / data.flux_pulse_amplitude
        ).tolist()
        if not data.has_filters(qubit):
            # Derive IIR
            acquisition_time = len(x)
            exp_params = exponential_params(step_response[qubit], acquisition_time)
            feedback_taps[qubit], feedforward_taps_iir[qubit] = filter_calc(exp_params)
            time_decay[qubit], alpha[qubit], g[qubit] = exp_params
            iir_correction = lfilter(
                feedforward_taps_iir[qubit], feedback_taps[qubit], step_response[qubit]
            )
            # FIR corrections

            taps = data.fir
            baseline = g[qubit]
            x0 = [1] + (taps - 1) * [0]

            def fir_cost_function(x):
                yc = lfilter(x, 1, iir_correction)
                return np.mean(np.abs(yc - baseline)) / np.abs(baseline)

            fir = cma.fmin2(fir_cost_function, x0, 0.5)[0]

            feedforward_taps[qubit] = np.convolve(
                feedforward_taps_iir[qubit], fir
            ).tolist()

            if np.max(np.abs(feedforward_taps[qubit])) > FEEDFORWARD_MAX:
                feedforward_taps[qubit] = (
                    2
                    * np.array(feedforward_taps[qubit])
                    / abs(max(feedforward_taps[qubit]))
                ).tolist()

    return CryoscopeResults(
        amplitude=amplitude,
        detuning=detuning,
        step_response=step_response,
        fitted_parameters=fitted_parameters,
        exp_amplitude=alpha,
        tau=time_decay,
        feedforward_taps=feedforward_taps,
        feedforward_taps_iir=feedforward_taps_iir,
        feedback_taps=feedback_taps,
    )


def _plot(data: CryoscopeData, fit: CryoscopeResults, target: QubitId):
    """Cryoscope plots."""

    fig = make_subplots(
        rows=2,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )
    duration = data[(target, "MX")].duration

    fig.add_trace(
        go.Scatter(
            x=duration,
            y=1 - 2 * data[(target, "MX")].prob_1,
            name="X",
            legendgroup="1",
        ),
        row=1,
        col=1,
    )
    fig.add_trace(
        go.Scatter(
            x=duration,
            y=1 - 2 * data[(target, "MY")].prob_1,
            name="Y",
            legendgroup="1",
        ),
        row=1,
        col=1,
    )

    fitting_report = ""
    if fit is not None:
        fig.add_trace(
            go.Scatter(
                x=duration,
                y=fit.step_response[target],
                name="Uncorrected waveform",
                legendgroup="2",
            ),
            row=2,
            col=1,
        )

        if not data.has_filters(target):
            iir_corrections = lfilter(
                fit.feedforward_taps_iir[target],
                fit.feedback_taps[target],
                fit.step_response[target],
            )
            all_corrections = lfilter(
                fit.feedforward_taps[target],
                fit.feedback_taps[target],
                fit.step_response[target],
            )

            fig.add_trace(
                go.Scatter(
                    x=duration,
                    y=iir_corrections,
                    name="IIR corrections",
                    legendgroup="2",
                ),
                row=2,
                col=1,
            )

            fig.add_trace(
                go.Scatter(
                    x=duration,
                    y=all_corrections,
                    name="FIR + IIR corrections",
                    legendgroup="2",
                ),
                row=2,
                col=1,
            )

            exp_amplitude = fit.exp_amplitude[target]
            tau = fit.tau[target]

            fitting_report = table_html(
                table_dict(
                    target,
                    ["A", "tau"],
                    [
                        exp_amplitude,
                        tau,
                    ],
                )
            )

        fig.update_layout(
            showlegend=True,
            legend_tracegroupgap=120,
            xaxis2_title="Duration [ns]",
            yaxis1_title="Expectation value",
            yaxis2_title="Waveform",
        )

    return [fig], fitting_report


def _update(results: CryoscopeResults, platform: Platform, target: QubitId):
    try:
        update.feedforward(results.feedforward_taps[target], platform, target)
        update.feedback(results.feedback_taps[target], platform, target)
    except KeyError:
        log.info(f"Skipping filters update on qubit {target}.")


cryoscope = Routine(_acquisition, _fit, _plot, _update)



from dataclasses import dataclass

import numpy as np
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.result import probability

from . import t1
from .spin_echo import SpinEchoParameters, SpinEchoResults
from .utils import dynamical_decoupling_sequence, exponential_fit_probability, plot


@dataclass
class CpmgParameters(SpinEchoParameters):
    """Cpmg runcard inputs."""

    n: int = 1
    """Number of pi rotations."""


@dataclass
class CpmgResults(SpinEchoResults):
    """SpinEcho outputs."""


class CpmgData(t1.T1Data):
    """SpinEcho acquisition outputs."""


def _acquisition(
    params: CpmgParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> CpmgData:
    """Data acquisition for Cpmg"""
    # create a sequence of pulses for the experiment:
    sequence, delays = dynamical_decoupling_sequence(
        platform, targets, n=params.n, kind="CPMG"
    )

    # define the parameter to sweep and its range:
    # delay between pulses
    wait_range = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )
    durations = []
    for q in targets:
        # this is assuming that RX and RX90 have the same duration
        duration = platform.natives.single_qubit[q].RX()[0][1].duration
        durations.append(duration)
        assert (
            params.delay_between_pulses_start - params.n * duration
        ) / 2 / params.n >= 0, (
            f"Initial delay too short for qubit {q}, "
            f"minimum delay should be {params.n * duration}"
        )

    assert len(set(durations)) == 1, (
        "Cannot run on mulitple qubit with different RX duration."
    )

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=(wait_range - params.n * durations[0]) / 2 / params.n,
        pulses=delays,
    )

    results = platform.execute(
        [sequence],
        [[sweeper]],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )

    data = CpmgData()
    for qubit in targets:
        ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[-1]
        result = results[ro_pulse.id]
        prob = probability(result, state=1)
        error = np.sqrt(prob * (1 - prob) / params.nshots)
        data.register_qubit(
            t1.CoherenceProbType,
            (qubit),
            dict(
                wait=wait_range,
                prob=prob,
                error=error,
            ),
        )

    return data


def _fit(data: CpmgData) -> CpmgResults:
    """Post-processing for Cpmg."""
    t2Echos, fitted_parameters, pcovs, chi2 = exponential_fit_probability(data)
    return CpmgResults(t2Echos, fitted_parameters, pcovs, chi2)


cpmg = Routine(_acquisition, _fit, plot)
"""Cpmg Routine object."""



from typing import Optional

from qibo import Circuit, gates
from qibo.backends import Backend
from qibo.transpiler.pipeline import Passes
from qibo.transpiler.unroller import NativeGates, Unroller
from qibolab import PulseSequence
from qibolab._core.compilers import Compiler
from qibolab._core.native import NativeContainer

from qibocal.auto.operation import QubitId

REPLACEMENTS = {
    "RX": "GPI2",
    "MZ": "M",
}


def transpile_circuits(
    circuits: list[Circuit],
    qubit_maps: list[list[QubitId]],
    backend: Backend,
    transpiler: Optional[Passes],
):
    """Transpile and pad `circuits` according to the platform.

    Apply the `transpiler` to `circuits` and pad them in
    circuits with the same number of qubits in the platform.
    Before manipulating the circuits, this function check that the
    `qubit_maps` contain string ids and in the positive case it
    remap them in integers, following the ids order provided by the
    platform.

    .. note::

        In this function we are implicitly assume that the qubit ids
        are all string or all integers.
    """
    transpiled_circuits = []
    platform = backend.platform
    qubits = list(platform.qubits)
    if isinstance(qubit_maps[0][0], str):
        for i, qubit_map in enumerate(qubit_maps):
            qubit_map = map(lambda x: qubits.index(x), qubit_map)
            qubit_maps[i] = list(qubit_map)
    platform_nqubits = platform.nqubits
    for circuit, qubit_map in zip(circuits, qubit_maps):
        new_circuit = pad_circuit(platform_nqubits, circuit, qubit_map)
        transpiled_circ, _ = transpiler(new_circuit)
        transpiled_circuits.append(transpiled_circ)

    return transpiled_circuits


def execute_transpiled_circuits(
    circuits: list[Circuit],
    qubit_maps: list[list[QubitId]],
    backend: Backend,
    transpiler: Optional[Passes],
    initial_states=None,
    nshots=1000,
):
    """Transpile `circuits`.

    If the `qibolab` backend is used, this function pads the `circuits` in new
    ones with a number of qubits equal to the one provided by the platform.
    At the end, the circuits are transpiled, executed and the results returned.
    The input `transpiler` is optional, but it should be provided if the backend
    is `qibolab`.
    For the qubit map look :func:`dummy_transpiler`.
    This function returns the list of transpiled circuits and the execution results.
    """
    transpiled_circuits = transpile_circuits(
        circuits,
        qubit_maps,
        backend,
        transpiler,
    )
    return transpiled_circuits, backend.execute_circuits(
        transpiled_circuits, initial_states=initial_states, nshots=nshots
    )


def execute_transpiled_circuit(
    circuit: Circuit,
    qubit_map: list[QubitId],
    backend: Backend,
    transpiler: Optional[Passes],
    initial_state=None,
    nshots=1000,
):
    """Transpile `circuit`.

    If the `qibolab` backend is used, this function pads the `circuit` in new a
    one with a number of qubits equal to the one provided by the platform.
    At the end, the circuit is transpiled, executed and the results returned.
    The input `transpiler` is optional, but it should be provided if the backend
    is `qibolab`.
    For the qubit map look :func:`dummy_transpiler`.
    This function returns the transpiled circuit and the execution results.
    """

    transpiled_circ = transpile_circuits(
        [circuit],
        [qubit_map],
        backend,
        transpiler,
    )[0]
    return transpiled_circ, backend.execute_circuit(
        transpiled_circ, initial_state=initial_state, nshots=nshots
    )


def natives(platform) -> dict[str, NativeContainer]:
    """
    Return the dict of native gates name with the associated native container
    defined in the `platform`. This function assumes the native gates to be the same for each
    qubit and pair.
    """
    qubit = next(iter(platform.qubits))
    single_qubit_natives_container = platform.natives.single_qubit[qubit]
    single_qubit_natives = list(single_qubit_natives_container.model_fields)
    if len(platform.pairs) > 0:
        # add two qubit natives only if there are pairs
        pair = next(iter(platform.pairs))
        two_qubit_natives_container = platform.natives.two_qubit[pair]
        two_qubit_natives = list(two_qubit_natives_container.model_fields)
    else:
        two_qubit_natives = []
    # Solve Qibo-Qibolab mismatch
    single_qubit_natives.append("RZ")
    single_qubit_natives.append("Z")
    single_qubit_natives.remove("RX12")
    single_qubit_natives.remove("RX90")
    single_qubit_natives.remove("CP")
    single_qubit_natives = [REPLACEMENTS.get(x, x) for x in single_qubit_natives]
    return {i: platform.natives.single_qubit[qubit] for i in single_qubit_natives} | {
        i: platform.natives.two_qubit[pair] for i in two_qubit_natives
    }


def create_rule(name, natives):
    """Create rule for gate name given container natives."""

    def rule(gate: gates.Gate, natives: NativeContainer) -> PulseSequence:
        return natives.ensure(name).create_sequence()

    return rule


def set_compiler(backend, natives_):
    """
    Set the compiler to execute the native gates defined by the platform.
    """
    compiler = backend.compiler
    rules = {}
    for name, natives_container in natives_.items():
        gate = getattr(gates, name)
        if gate not in compiler.rules:
            rules[gate] = create_rule(name, natives_container)
        else:
            rules[gate] = compiler.rules[gate]
    rules[gates.I] = compiler.rules[gates.I]
    backend.compiler = Compiler(rules=rules)


def dummy_transpiler(backend: Backend) -> Passes:
    """
    If the backend is `qibolab`, a transpiler with just an unroller is returned,
    otherwise `None`. This function overwrites the compiler defined in the
    backend, taking into account the native gates defined in the`platform` (see
    :func:`set_compiler`).
    """
    platform = backend.platform
    native_gates = natives(platform)
    set_compiler(backend, native_gates)
    native_gates = [getattr(gates, x) for x in native_gates]
    unroller = Unroller(NativeGates.from_gatelist(native_gates))
    return Passes(connectivity=platform.pairs, passes=[unroller])


def pad_circuit(nqubits, circuit: Circuit, qubit_map: list[int]) -> Circuit:
    """
    Pad `circuit` in a new one with `nqubits` qubits, according to `qubit_map`.
    `qubit_map` is a list `[i, j, k, ...]`, where the i-th physical qubit is mapped
    into the 0th logical qubit and so on.
    """
    new_circuit = Circuit(nqubits)
    new_circuit.add(circuit.on_qubits(*qubit_map))
    return new_circuit



import json
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibo import Circuit, gates
from qibo.backends import NumpyBackend, construct_backend, matrices
from qibo.quantum_info import fidelity, partial_trace

from qibocal.auto.operation import DATAFILE, Data, Parameters, QubitId, Results, Routine
from qibocal.auto.transpile import dummy_transpiler, execute_transpiled_circuit
from qibocal.calibration import CalibrationPlatform

from .utils import table_dict, table_html

BASIS = ["X", "Y", "Z"]
"""Single qubit measurement basis."""
CIRCUIT_PATH = "circuit.json"
"""Path where circuit is stored."""


@dataclass
class StateTomographyParameters(Parameters):
    """Tomography input parameters"""

    circuit: Optional[Union[str, Circuit]] = None
    """Circuit to prepare initial state.

        It can also be provided the path to a json file containing
        a serialized circuit.
    """

    def __post_init__(self):
        if isinstance(self.circuit, str):
            raw = json.loads((Path.cwd() / self.circuit).read_text())
            self.circuit = Circuit.from_dict(raw)


TomographyType = np.dtype(
    [
        ("samples", np.int64),
    ]
)
"""Custom dtype for tomography."""


@dataclass
class StateTomographyData(Data):
    """Tomography data"""

    targets: dict[QubitId, int]
    """Store targets order."""
    circuit: Circuit
    """Circuit where tomography will be executed."""
    data: dict[tuple[QubitId, str], np.int64] = field(default_factory=dict)
    """Hardware measurements."""

    @property
    def params(self) -> dict:
        """Convert non-arrays attributes into dict."""
        params = super().params
        params.pop("circuit")
        return params

    def save(self, path):
        super().save(path)
        (path / CIRCUIT_PATH).write_text(json.dumps(self.circuit.raw))

    @classmethod
    def load(cls, path):
        circuit = Circuit.from_dict(json.loads((path / CIRCUIT_PATH).read_text()))
        data = super().load_data(path, DATAFILE)
        params = super().load_params(path, DATAFILE)
        return cls(data=data, circuit=circuit, targets=params["targets"])


@dataclass
class StateTomographyResults(Results):
    """Tomography results"""

    measured_density_matrix_real: dict[QubitId, list]
    """Real part of measured density matrix."""
    measured_density_matrix_imag: dict[QubitId, list]
    """Imaginary part of measured density matrix."""
    target_density_matrix_real: dict[QubitId, list]
    """Real part of exact density matrix."""
    target_density_matrix_imag: dict[QubitId, list]
    """Imaginary part of exact density matrix."""
    fidelity: dict[QubitId, float]
    """State fidelity."""


def _acquisition(
    params: StateTomographyParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> StateTomographyData:
    """Acquisition protocol for single qubit state tomography experiment."""
    if params.circuit is None:
        params.circuit = Circuit(len(targets))

    backend = construct_backend("qibolab", platform=platform)
    transpiler = dummy_transpiler(backend)

    data = StateTomographyData(
        circuit=params.circuit, targets={target: i for i, target in enumerate(targets)}
    )

    for basis in BASIS:
        basis_circuit = deepcopy(params.circuit)
        # FIXME: https://github.com/qiboteam/qibo/issues/1318
        if basis != "Z":
            for i in range(len(targets)):
                basis_circuit.add(getattr(gates, basis)(i).basis_rotation())

        basis_circuit.add(gates.M(*range(len(targets))))
        _, results = execute_transpiled_circuit(
            basis_circuit,
            targets,
            backend,
            nshots=params.nshots,
            transpiler=transpiler,
        )
        for i, target in enumerate(targets):
            data.register_qubit(
                TomographyType,
                (target, basis),
                dict(
                    samples=np.array(results.samples()).T[i],
                ),
            )
    return data


def _fit(data: StateTomographyData) -> StateTomographyResults:
    """Post-processing for State tomography."""
    measured_density_matrix_real = {}
    measured_density_matrix_imag = {}
    target_density_matrix_real = {}
    target_density_matrix_imag = {}
    fid = {}
    circuit = data.circuit
    circuit.density_matrix = True
    total_density_matrix = NumpyBackend().execute_circuit(circuit=circuit).state()
    for i, qubit in enumerate(data.targets):
        traced_qubits = [q for q in range(len(data.qubits)) if q != i]
        target_density_matrix = partial_trace(total_density_matrix, traced_qubits)
        x_exp = 1 - 2 * np.mean(data[qubit, "X"].samples)
        y_exp = 1 - 2 * np.mean(data[qubit, "Y"].samples)
        z_exp = 1 - 2 * np.mean(data[qubit, "Z"].samples)
        measured_density_matrix = 0.5 * (
            matrices.I + matrices.X * x_exp + matrices.Y * y_exp + matrices.Z * z_exp
        )
        measured_density_matrix_real[qubit] = np.real(measured_density_matrix).tolist()
        measured_density_matrix_imag[qubit] = np.imag(measured_density_matrix).tolist()

        target_density_matrix_real[qubit] = np.real(target_density_matrix).tolist()
        target_density_matrix_imag[qubit] = np.imag(target_density_matrix).tolist()
        fid[qubit] = fidelity(
            measured_density_matrix,
            target_density_matrix,
        )

    return StateTomographyResults(
        measured_density_matrix_real=measured_density_matrix_real,
        measured_density_matrix_imag=measured_density_matrix_imag,
        target_density_matrix_real=target_density_matrix_real,
        target_density_matrix_imag=target_density_matrix_imag,
        fidelity=fid,
    )


def plot_parallelogram(a, e, pos_x, pos_y, **options):
    """Plotting single histogram in 3d plot."""
    x, y, z = np.meshgrid(
        np.linspace(pos_x - a / 4, pos_x + a / 4, 2),
        np.linspace(pos_y - a / 4, pos_y + a / 4, 2),
        np.linspace(0, e, 2),
    )
    x = x.flatten()
    y = y.flatten()
    z = z.flatten()
    return go.Mesh3d(
        x=x,
        y=y,
        z=z,
        alphahull=1,
        flatshading=True,
        lighting={"diffuse": 0.1, "specular": 2.0, "roughness": 0.5},
        **options,
    )


def plot_rho(fig, zz, trace_options, figure_options, showlegend=None):
    """Plot density matrix"""
    values = list(range(len(zz)))
    x, y = np.meshgrid(values, values)
    xx = x.flatten()
    yy = y.flatten()
    zz = np.array(zz).ravel()
    showlegend_temp = False
    for x, y, z in zip(xx, yy, zz):
        if showlegend is None:
            showlegend_temp = bool(x == xx[-1] and y == yy[-1])
        fig.add_trace(
            plot_parallelogram(1, z, x, y, showlegend=showlegend_temp, **trace_options),
            **figure_options,
        )


def plot_reconstruction(ideal, measured):
    """Plot 3D plot with reconstruction of ideal and measured density matrix."""
    fig = make_subplots(
        rows=1,
        cols=2,
        start_cell="top-left",
        specs=[[{"type": "scatter3d"}, {"type": "scatter3d"}]],
        subplot_titles=(
            "Re(ρ)",
            "Im(ρ)",
        ),
    )

    # computing limits for colorscale
    min_re, max_re = np.min(ideal.real), np.max(ideal.real)
    min_im, max_im = np.min(ideal.imag), np.max(ideal.imag)

    # add offset
    if np.abs(min_re - max_re) < 1e-5:
        min_re = min_re - 0.1
        max_re = max_re + 0.1
    if np.abs(min_im - max_im) < 1e-5:
        min_im = min_im - 0.1
        max_im = max_im + 0.1

    plot_rho(
        fig,
        measured.real,
        trace_options=dict(
            color="rgba(255,100,0,0.1)", name="experiment", legendgroup="experiment"
        ),
        figure_options=dict(row=1, col=1),
    )

    plot_rho(
        fig,
        ideal.real,
        trace_options=dict(
            color="rgba(100,0,100,0.1)", name="simulation", legendgroup="simulation"
        ),
        figure_options=dict(row=1, col=1),
    )

    plot_rho(
        fig,
        measured.imag,
        trace_options=dict(
            color="rgba(255,100,0,0.1)", name="experiment", legendgroup="experiment"
        ),
        figure_options=dict(row=1, col=2),
        showlegend=False,
    )

    plot_rho(
        fig,
        ideal.imag,
        trace_options=dict(
            color="rgba(100,0,100,0.1)", name="simulation", legendgroup="simulation"
        ),
        figure_options=dict(row=1, col=2),
        showlegend=False,
    )

    tickvals = list(range(len(ideal)))
    if len(tickvals) == 2:  # single qubit tomography
        ticktext = ["{:01b}".format(i) for i in tickvals]
    else:  # two qubit tomography
        ticktext = ["{:02b}".format(i) for i in tickvals]
    fig.update_scenes(
        xaxis=dict(tickvals=tickvals, ticktext=ticktext),
        yaxis=dict(tickvals=tickvals, ticktext=ticktext),
        zaxis=dict(range=[-1, 1]),
    )

    return fig


def _plot(data: StateTomographyData, fit: StateTomographyResults, target: QubitId):
    """Plotting for state tomography"""
    if fit is None:
        return [], ""

    ideal = np.array(fit.target_density_matrix_real[target]) + 1j * np.array(
        fit.target_density_matrix_imag[target]
    )
    measured = np.array(fit.measured_density_matrix_real[target]) + 1j * np.array(
        fit.measured_density_matrix_imag[target]
    )
    fig = plot_reconstruction(ideal, measured)

    fitting_report = table_html(
        table_dict(
            target,
            [
                "Fidelity",
            ],
            [
                np.round(fit.fidelity[target], 4),
            ],
        )
    )

    return [fig], fitting_report


state_tomography = Routine(_acquisition, _fit, _plot)



import json
import pathlib

import numpy as np
from qibo import gates
from qibo.models import Circuit

SINGLE_QUBIT_CLIFFORDS_NAMES = {
    # Virtual gates
    "": gates.I,
    "minusX,minusY": lambda q: gates.U3(q, 0, np.pi / 2, np.pi / 2),  # Z
    "sqrtX,sqrtMinusY,sqrtMinusX": lambda q: gates.U3(
        q, 0, -np.pi / 2, 0
    ),  # gates.RZ(q, np.pi / 2)
    "sqrtX,sqrtY,sqrtMinusX": lambda q: gates.U3(
        q, 0, np.pi / 2, 0
    ),  # gates.RZ(q, -np.pi / 2)
    # pi rotations
    "minusX": lambda q: gates.U3(q, np.pi, -np.pi, 0),  # X
    "minusY": lambda q: gates.U3(q, np.pi, 0, 0),  # Y
    # pi/2 rotations
    "sqrtX": lambda q: gates.U3(q, np.pi / 2, -np.pi / 2, np.pi / 2),  # Rx(pi/2)
    "sqrtMinusX": lambda q: gates.U3(q, -np.pi / 2, -np.pi / 2, np.pi / 2),  # Rx(-pi/2)
    "sqrtY": lambda q: gates.U3(q, np.pi / 2, 0, 0),  # Ry(pi/2)
    "sqrtMinusY": lambda q: gates.U3(q, -np.pi / 2, 0, 0),  # Ry(-pi/2)
    # 2pi/3 rotations
    "sqrtX,sqrtY": lambda q: gates.U3(q, np.pi / 2, -np.pi / 2, 0),  # Rx(pi/2)Ry(pi/2)
    "sqrtX,sqrtMinusY": lambda q: gates.U3(
        q, np.pi / 2, -np.pi / 2, np.pi
    ),  # Rx(pi/2)Ry(-pi/2)
    "sqrtMinusX,sqrtY": lambda q: gates.U3(
        q, np.pi / 2, np.pi / 2, 0
    ),  # Rx(-pi/2)Ry(pi/2)
    "sqrtMinusX,sqrtMinusY": lambda q: gates.U3(
        q, np.pi / 2, np.pi / 2, -np.pi
    ),  # Rx(-pi/2)Ry(-pi/2)
    "sqrtY,sqrtX": lambda q: gates.U3(q, np.pi / 2, 0, np.pi / 2),  # Ry(pi/2)Rx(pi/2)
    "sqrtY,sqrtMinusX": lambda q: gates.U3(
        q, np.pi / 2, 0, -np.pi / 2
    ),  # Ry(pi/2)Rx(-pi/2)
    "sqrtMinusY,sqrtX": lambda q: gates.U3(
        q, np.pi / 2, -np.pi, np.pi / 2
    ),  # Ry(-pi/2)Rx(pi/2)
    "sqrtMinusY,sqrtMinusX": lambda q: gates.U3(
        q, np.pi / 2, np.pi, -np.pi / 2
    ),  # Ry(-pi/2)Rx(-pi/2)
    # Hadamard-like
    "minusX,sqrtY": lambda q: gates.U3(q, np.pi / 2, -np.pi, 0),  # X Ry(pi/2)
    "minusX,sqrtMinusY": lambda q: gates.U3(q, np.pi / 2, 0, np.pi),  # X Ry(-pi/2)
    "minusY,sqrtX": lambda q: gates.U3(
        q, np.pi / 2, np.pi / 2, np.pi / 2
    ),  # Y Rx(pi/2)
    "minusY,sqrtMinusX": lambda q: gates.U3(
        q, np.pi / 2, -np.pi / 2, -np.pi / 2
    ),  # Y Rx(-pi/2)
    "sqrtX,sqrtY,sqrtX": lambda q: gates.U3(
        q, np.pi, -np.pi / 4, np.pi / 4
    ),  # Rx(pi/2)Ry(pi/2)Rx(pi/2)
    "sqrtX,sqrtMinusY,sqrtX": lambda q: gates.U3(
        q, np.pi, np.pi / 4, -np.pi / 4
    ),  # Rx(-pi/2)Ry(pi/2)Rx(-pi/2)
}


# TODO: Expand when more entangling gates are calibrated
def find_cliffords(cz_list):
    """Splits a clifford (list of gates) into sublists based on the occurrence of the "CZ" gate."""
    clifford_list = []
    clifford = []
    for gate in cz_list:
        if gate == "CZ":
            clifford.append(gate)
            clifford_list.append(clifford)
            clifford = []
            continue
        clifford.append(gate)
    clifford_list.append(clifford)
    return clifford_list


def separator(clifford):
    """
    Separates values in the given clifford sublist based on certain conditions.

    Returns:
        tuple: A tuple containing three elements:
            - values_with_1 (str): A comma-separated string of values containing '1'.
            - values_with_2 (str): A comma-separated string of values containing '2'.
            - value_with_CZ (bool): True if 'CZ' is present in the clifford list, False otherwise.
    """

    # Separate values containing 1
    values_with_1 = [value for value in clifford if "1" in value]
    values_with_1 = ",".join(values_with_1)

    # Separate values containing 2
    values_with_2 = [value for value in clifford if "2" in value]
    values_with_2 = ",".join(values_with_2)

    # Check if CZ
    value_with_CZ = [value for value in clifford if "CZ" in value]
    value_with_CZ = len(value_with_CZ) == 1  # FIXME: What is this ?

    values_with_1 = values_with_1.replace("1", "")
    values_with_2 = values_with_2.replace("2", "")
    return values_with_1, values_with_2, value_with_CZ


def clifford2gates(clifford):
    """
    Converts a Clifford string into a list of gates.

    Args:
        clifford (str): A comma-separated string representing a sequence of gates that represent a Clifford gate.
    """
    gate_list = clifford.split(",")

    clifford_list = find_cliffords(gate_list)

    clifford_gate = []
    for clifford in clifford_list:
        values_with_1, values_with_2, value_with_CZ = separator(clifford)
        clifford_gate.append(SINGLE_QUBIT_CLIFFORDS_NAMES[values_with_1](0))
        clifford_gate.append(SINGLE_QUBIT_CLIFFORDS_NAMES[values_with_2](1))
        if value_with_CZ:
            clifford_gate.append(gates.CZ(0, 1))

    return clifford_gate


def clifford_to_matrix(clifford):
    """
    Converts a Clifford gate as a string to its corresponding unitary matrix representation.
    """
    clifford_gate = clifford2gates(clifford)

    qubits_str = ["q0", "q1"]

    new_circuit = Circuit(2, wire_names=qubits_str)
    for gate in clifford_gate:
        new_circuit.add(gate)

    unitary = new_circuit.unitary()

    return unitary


def generate_inv_dict_cliffords_file(two_qubit_cliffords, output_file=None):
    """
    Generate an inverse dictionary of Clifford matrices and save it to a npz file.

    Parameters:
    two_qubit_cliffords (dict): A dictionary of two-qubit Cliffords.
    output_file (str): The path to the output npz file.
    """
    clifford_matrices = {}
    for i, clifford in enumerate(two_qubit_cliffords.values()):
        clifford = two_qubit_cliffords[str(i)]

        unitary = clifford_to_matrix(clifford)
        unitary = unitary.round(3)
        unitary += 0.0 + 0.0j

        clifford_matrices[i] = unitary

    clifford_matrices_inv_np = {}
    # Convert the arrays to strings and store them as keys in the new dictionary
    for key, value in clifford_matrices.items():
        key_str = np.array2string(value, separator=",")
        clifford_matrices_inv_np[key_str] = key

    if output_file is not None:
        np.savez(output_file, **clifford_matrices_inv_np)

    return clifford_matrices_inv_np


def clifford_to_pulses(clifford):
    """
    From a Clifford gate sequence into the number of pulses required to implement it.

    Args:
        clifford (str): A comma-separated string representing the Clifford gate sequence.

    Returns:
        int: The number of pulses required to implement the given Clifford gate sequence.
    """
    gate_list = clifford.split(",")

    clifford_list = find_cliffords(gate_list)

    pulses = 0
    for clifford in clifford_list:
        values_with_1, values_with_2, value_with_CZ = separator(clifford)

        if SINGLE_QUBIT_CLIFFORDS_NAMES[values_with_1](0).name != "id":
            pulses += 2  # This assumes a U3 transpiled into 2 pulses
        if SINGLE_QUBIT_CLIFFORDS_NAMES[values_with_2](1).name != "id":
            pulses += 2  # This assumes a U3 transpiled into 2 pulses
        if value_with_CZ:
            pulses += 1  # This assumes a CZ without parking so 1 pulse

    return pulses


def calculate_pulses_clifford(cliffords):
    """
    Calculate the average number of pulses per Clifford operation.

    Parameters:
    - cliffords (dict): A dictionary of Clifford operations.

    Returns:
    - pulses_per_clifford (float): The average number of pulses per Clifford operation.
    """
    pulses = 0
    for i, clifford in enumerate(cliffords.values()):
        clifford = cliffords[str(i)]
        pulses += clifford_to_pulses(clifford)

    pulses_per_clifford = pulses / len(cliffords)
    return pulses_per_clifford


def load_inverse_cliffords(file_inv):
    path = pathlib.Path(__file__).parent / file_inv
    clifford_matrices_inv = np.load(path)
    return clifford_matrices_inv


def load_cliffords(file_cliffords):
    path = pathlib.Path(__file__).parent / file_cliffords
    with open(path) as file:
        two_qubit_cliffords = json.load(file)
    return two_qubit_cliffords



"""Common result operations."""

import numpy as np
import numpy.typing as npt

IQ = npt.NDArray[np.float64]
"""An array of I and Q values.

It is assumed that the I and Q component are discriminated by the
innermost dimension of the array.
"""


def _lift(values: IQ) -> npt.NDArray:
    """Transpose the innermost dimension to the outermost."""
    return np.moveaxis(values, -1, 0)


def _sink(values: npt.NDArray) -> IQ:
    """Transpose the outermost dimension to the innermost.

    Inverse of :func:`_lift`.
    """
    return np.moveaxis(values, 0, -1)


def collect(i: npt.NDArray, q: npt.NDArray) -> IQ:
    """Collect I and Q components in a single array."""
    return _sink(np.stack([i, q]))


def unpack(iq: IQ) -> tuple[npt.NDArray, npt.NDArray]:
    """Unpack I and Q components from single array.

    Inverse of :func:`collect`.
    """
    i, q = tuple(_lift(iq))
    return i, q


def magnitude(iq: IQ):
    """Signal magnitude.

    It is supposed to be a tension, possibly in arbitrary units.
    """
    iq_ = _lift(iq)
    return np.sqrt(iq_[0] ** 2 + iq_[1] ** 2)


def average(values: npt.NDArray) -> tuple[npt.NDArray, npt.NDArray]:
    """Perform the values average.

    It returns both the average estimator itself, and its standard
    deviation estimator.

    Use this also for I and Q values in the *standard layout*, cf. :class:`IQ`.
    """
    mean = np.mean(values, axis=0)
    std = np.std(values, axis=0, ddof=1) / np.sqrt(values.shape[0])
    return mean, std


def average_iq(i: npt.NDArray, q: npt.NDArray) -> tuple[npt.NDArray, npt.NDArray]:
    """Perform the average over I and Q.

    Convenience wrapper over :func:`average` for separate i and q samples arrays.
    """
    return average(collect(i, q))


def phase(iq: npt.NDArray):
    """Signal phase in radians.

    It is assumed that the I and Q component are discriminated by the
    innermost dimension of the array.
    """
    iq_ = _lift(iq)
    return np.unwrap(np.arctan2(iq_[0], iq_[1]))


def probability(values: npt.NDArray, state: int = 0):
    """Return the statistical frequency of the specified state."""
    return np.sum(values == state, axis=0) / len(values)



from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibolab import AcquisitionType, AveragingMode, Parameter, PulseSequence, Sweeper

from qibocal import update
from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.protocols.utils import (
    HZ_TO_GHZ,
    lorentzian,
    lorentzian_fit,
    readout_frequency,
    table_dict,
    table_html,
)

from ..result import magnitude, phase, unpack


@dataclass
class DispersiveShiftParameters(Parameters):
    """Dispersive shift inputs."""

    freq_width: int
    """Width [Hz] for frequency sweep relative to the readout frequency [Hz]."""
    freq_step: int
    """Frequency step for sweep [Hz]."""


@dataclass
class DispersiveShiftResults(Results):
    """Dispersive shift outputs."""

    frequencies: dict[QubitId, list[float]]
    """Qubit peak frequencies."""
    fitted_parameters: dict[QubitId, list[list[float]]]
    """Fitted parameters. The first element is the resonator frequency when the
    qubit is in the ground state, the second one when the qubit is in the first excited
    state."""
    best_freq: dict[QubitId, float]
    """Readout frequency that maximizes the distance of ground and excited states in iq-plane"""

    def chi(self, target: QubitId) -> float:
        "Evaluate the dispersive shift"
        freq = self.frequencies[target]
        return (freq[0] - freq[1]) / 2


DispersiveShiftType = np.dtype(
    [
        ("freq", np.float64),
        ("i", np.float64),
        ("q", np.float64),
        ("signal", np.float64),
        ("phase", np.float64),
    ]
)
"""Custom dtype for dispersive shift."""


@dataclass
class DispersiveShiftData(Data):
    """Dispersive shift acquisition outputs."""

    resonator_type: str
    """Resonator type."""
    data: dict[tuple[QubitId, int], npt.NDArray[DispersiveShiftType]] = field(
        default_factory=dict
    )


def _acquisition(
    params: DispersiveShiftParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> DispersiveShiftData:
    r"""
    Data acquisition for dispersive shift experiment.
    Perform spectroscopy on the readout resonator, with the qubit in ground and excited state, showing
    the resonator shift produced by the coupling between the resonator and the qubit.

    Args:
        params (DispersiveShiftParameters): experiment's parameters
        platform (CalibrationPlatform): Qibolab platform object
        targets (list): list of target qubits to perform the action
    """

    sequence_0 = PulseSequence()
    sequence_1 = PulseSequence()
    for qubit in targets:
        natives = platform.natives.single_qubit[qubit]
        sequence_0 += natives.MZ()
        sequence_1 += natives.RX() | natives.MZ()

    delta_frequency_range = np.arange(
        -params.freq_width / 2, params.freq_width / 2, params.freq_step
    )

    data = DispersiveShiftData(resonator_type=platform.resonator_type)

    sweepers = [
        Sweeper(
            parameter=Parameter.frequency,
            values=readout_frequency(q, platform) + delta_frequency_range,
            channels=[platform.qubits[q].probe],
        )
        for q in targets
    ]

    results = platform.execute(
        [sequence_0, sequence_1],
        [sweepers],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=AveragingMode.CYCLIC,
    )

    for qubit in targets:
        for state, sequence in enumerate([sequence_0, sequence_1]):
            ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[-1]
            result = results[ro_pulse.id]
            i, q = unpack(result)
            data.register_qubit(
                DispersiveShiftType,
                (qubit, state),
                dict(
                    freq=readout_frequency(qubit, platform) + delta_frequency_range,
                    signal=magnitude(result),
                    phase=phase(result),
                    i=i,
                    q=q,
                ),
            )
    return data


def _fit(data: DispersiveShiftData) -> DispersiveShiftResults:
    """Post-Processing for dispersive shift"""
    qubits = data.qubits
    iq_couples = [[], []]  # axis 0: states, axis 1: qubit

    res_frequencies = {}
    best_freqs = {}
    fitted_parameters = {}

    for qubit in qubits:
        freq = []
        fit_params = []
        for i in range(2):
            data_i = data[qubit, i]
            fit_result = lorentzian_fit(
                data_i, resonator_type=data.resonator_type, fit="resonator"
            )
            if fit_result is None:
                freq = fit_params = None
                break

            freq.append(fit_result[0])
            fit_params.append(fit_result[1])

        res_frequencies[qubit] = freq
        fitted_parameters[qubit] = fit_params

    for idx, qubit in enumerate(qubits):
        for i in range(2):
            data_i = data[qubit, i]
            i_measures = data_i.i
            q_measures = data_i.q
            iq_couples[i].append(np.stack((i_measures, q_measures), axis=-1))

        frequencies = data[qubit, 0].freq
        max_index = np.argmax(
            np.linalg.norm(iq_couples[0][idx] - iq_couples[1][idx], axis=-1)
        )
        best_freqs[qubit] = frequencies[max_index]

    return DispersiveShiftResults(
        frequencies=res_frequencies,
        fitted_parameters=fitted_parameters,
        best_freq=best_freqs,
    )


def _plot(data: DispersiveShiftData, target: QubitId, fit: DispersiveShiftResults):
    """Plotting function for dispersive shift."""
    figures = []
    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "Signal [a.u.]",
            "phase [rad]",
        ),
    )

    fitting_report = ""

    data_0 = data[target, 0]
    data_1 = data[target, 1]

    for i, label, q_data in list(
        zip(
            (0, 1),
            ("State 0", "State 1"),
            (data_0, data_1),
        )
    ):
        opacity = 1
        frequencies = q_data.freq * HZ_TO_GHZ
        fig.add_trace(
            go.Scatter(
                x=frequencies,
                y=q_data.signal,
                opacity=opacity,
                name=f"{label}",
                showlegend=True,
                legendgroup=f"{label}",
            ),
            row=1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=frequencies,
                y=q_data.phase,
                opacity=opacity,
                showlegend=False,
                legendgroup=f"{label}",
            ),
            row=1,
            col=2,
        )
        if fit is not None:
            fig.add_vline(
                x=fit.best_freq[target] * HZ_TO_GHZ,
                line=dict(color="orange", width=3, dash="dash"),
                row=1,
                col=1,
            )
            table_entries = [
                "Best Frequency [Hz]",
            ]
            table_values = np.round(
                [
                    fit.best_freq[target],
                ]
            )

            if fit.frequencies[target] is not None:
                freqrange = np.linspace(
                    min(frequencies),
                    max(frequencies),
                    2 * len(q_data),
                )
                params = fit.fitted_parameters[target][i]
                fig.add_trace(
                    go.Scatter(
                        x=freqrange,
                        y=lorentzian(freqrange, *params),
                        name=f"{label} Fit",
                        line=go.scatter.Line(dash="dot"),
                    ),
                    row=1,
                    col=1,
                )

                fig.add_trace(
                    go.Scatter(
                        x=[
                            fit.best_freq[target] * HZ_TO_GHZ,
                            fit.best_freq[target] * HZ_TO_GHZ,
                        ],
                        y=[
                            np.min(np.concatenate((data_0.signal, data_1.signal))),
                            np.max(np.concatenate((data_0.signal, data_1.signal))),
                        ],
                        mode="lines",
                        line=go.scatter.Line(color="orange", width=3, dash="dash"),
                        name="Best frequency",
                    ),
                    row=1,
                    col=1,
                )
                table_entries = [
                    "State Zero Frequency [Hz]",
                    "State One Frequency [Hz]",
                    "Chi [Hz]",
                    "Best Frequency [Hz]",
                ]
                table_values = np.round(
                    [
                        fit.frequencies[target][0],
                        fit.frequencies[target][1],
                        fit.chi(target),
                        fit.best_freq[target],
                    ]
                )

            fitting_report = table_html(table_dict(target, table_entries, table_values))
    fig.update_layout(
        showlegend=True,
        xaxis_title="Frequency [GHz]",
        yaxis_title="Signal [a.u.]",
        xaxis2_title="Frequency [GHz]",
        yaxis2_title="Phase [rad]",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(
    results: DispersiveShiftResults, platform: CalibrationPlatform, target: QubitId
):
    update.readout_frequency(results.best_freq[target], platform, target)
    if results.frequencies[target] is not None:
        delta = (
            platform.calibration.single_qubits[target].qubit.frequency_01
            - results.frequencies[target][0]
        )
        g = np.sqrt(np.abs(results.chi(target) * delta))
        update.coupling(g, platform, target)
        update.dressed_resonator_frequency(
            results.frequencies[target][0], platform, target
        )
        platform.calibration.single_qubits[target].readout.qudits_frequency[1] = (
            results.frequencies[target][1]
        )


dispersive_shift = Routine(_acquisition, _fit, _plot, _update)
"""Dispersive shift Routine object."""



from dataclasses import dataclass
from typing import Iterable, Optional, TypedDict, Union

import numpy as np
import plotly.graph_objects as go

from qibocal.auto.operation import Parameters, QubitId, Routine
from qibocal.calibration import CalibrationPlatform

from ..utils import table_dict, table_html
from .fitting import exp1B_func
from .utils import RBData, StandardRBResult, fit, number_to_str, rb_acquisition


class Depthsdict(TypedDict):
    """Dictionary used to build a list of depths as ``range(start, stop,
    step)``."""

    start: int
    stop: int
    step: int


@dataclass
class StandardRBParameters(Parameters):
    """Standard Randomized Benchmarking runcard inputs."""

    depths: Union[list, Depthsdict]
    """A list of depths/sequence lengths.

    If a dictionary is given the list will be build.
    """
    niter: int
    """Sets how many iterations over the same depth value."""
    uncertainties: Optional[float] = None
    """Method of computing the error bars of the signal and uncertainties of
    the fit.

    If ``None``,
    it computes the standard deviation. Otherwise it computes the corresponding confidence interval. Defaults `None`.
    """
    unrolling: bool = False
    """If ``True`` it uses sequence unrolling to deploy multiple circuits in a
    single instrument call.

    Defaults to ``False``.
    """
    seed: Optional[int] = None
    """A fixed seed to initialize ``np.random.Generator``.

    If ``None``, uses a random seed.
    Defaults is ``None``.
    """
    nshots: int = 10
    """Just to add the default value."""

    def __post_init__(self):
        if isinstance(self.depths, dict):
            self.depths = list(
                range(self.depths["start"], self.depths["stop"], self.depths["step"])
            )


def _acquisition(
    params: StandardRBParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> RBData:
    """The data acquisition stage of Standard Randomized Benchmarking.

    1. Set up the scan
    2. Execute the scan
    3. Post process the data and initialize a standard rb data object with it.

    Args:
        params: All parameters in one object.
        platform: CalibrationPlatform the experiment is executed on.
        target: list of qubits the experiment is executed on.

    Returns:
        RBData: The depths, samples and ground state probability of each experiment in the scan.
    """

    return rb_acquisition(params, platform, targets)


def _fit(data: RBData) -> StandardRBResult:
    """Takes a data frame, extracts the depths and the signal and fits it with
    an exponential function y = Ap^x+B.

    Args:
        data (RBData): Data from the data acquisition stage.

    Returns:
        StandardRBResult: Aggregated and processed data.
    """
    return fit(data.qubits, data)


def _plot(
    data: RBData, fit: StandardRBResult, target: QubitId
) -> tuple[list[go.Figure], str]:
    """Builds the table for the qq pipe, calls the plot function of the result
    object and returns the figure es list.

    Args:
        data (RBData): Data object used for the table.
        fit (StandardRBResult): Is called for the plot.
        target (_type_): Not used yet.

    Returns:
        tuple[list[go.Figure], str]:
    """
    target = tuple(target) if isinstance(target, list) else target
    qubit = target
    fig = go.Figure()
    fitting_report = ""
    x = data.depths
    raw_data = data.extract_probabilities(qubit)
    y = np.mean(raw_data, axis=1)
    raw_depths = [[depth] * data.niter for depth in data.depths]

    fig.add_trace(
        go.Scatter(
            x=np.hstack(raw_depths),
            y=np.hstack(raw_data),
            line=dict(color="#6597aa"),
            mode="markers",
            marker={"opacity": 0.2, "symbol": "square"},
            name="iterations",
        )
    )

    fig.add_trace(
        go.Scatter(
            x=x,
            y=y,
            line=dict(color="#aa6464"),
            mode="markers",
            name="average",
        )
    )
    # Create a dictionary for the error bars
    error_y_dict = None
    if fit is not None:
        popt, perr = fit.fit_parameters[qubit], fit.fit_uncertainties[qubit]
        label = "Fit: y=Ap^x<br>A: {}<br>p: {}<br>B: {}".format(
            number_to_str(popt[0], perr[0]),
            number_to_str(popt[1], perr[1]),
            number_to_str(popt[2], perr[2]),
        )
        x_fit = np.linspace(min(x), max(x), len(x) * 20)
        y_fit = exp1B_func(x_fit, *popt)
        fig.add_trace(
            go.Scatter(
                x=x_fit,
                y=y_fit,
                name=label,
                line=go.scatter.Line(dash="dot", color="#00cc96"),
            )
        )
        if fit.error_bars is not None:
            error_bars = fit.error_bars[qubit]
            # Constant error bars
            if isinstance(error_bars, Iterable) is False:
                error_y_dict = {"type": "constant", "value": error_bars}
            # Symmetric error bars
            elif isinstance(error_bars[0], Iterable) is False:
                error_y_dict = {"type": "data", "array": error_bars}
            # Asymmetric error bars
            else:
                error_y_dict = {
                    "type": "data",
                    "symmetric": False,
                    "array": error_bars[1],
                    "arrayminus": error_bars[0],
                }
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    error_y=error_y_dict,
                    line={"color": "#aa6464"},
                    mode="markers",
                    name="error bars",
                )
            )

    if fit is not None:
        fitting_report = table_html(
            table_dict(
                str(qubit),
                ["niter", "nshots", "uncertainties", "fidelity", "pulse_fidelity"],
                [
                    data.niter,
                    data.nshots,
                    data.uncertainties,
                    number_to_str(
                        fit.fidelity[qubit],
                        np.array(fit.fit_uncertainties[qubit][1]) / 2,
                    ),
                    number_to_str(
                        fit.pulse_fidelity[qubit],
                        np.array(fit.fit_uncertainties[qubit][1])
                        / (2 * data.npulses_per_clifford),
                    ),
                ],
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Circuit depth",
        yaxis_title="Survival Probability",
    )

    return [fig], fitting_report


def _update(results: StandardRBResult, platform: CalibrationPlatform, target: QubitId):
    """Write rb fidelity in calibration."""

    # TODO: shall we use the gate fidelity or the pulse fidelity
    platform.calibration.single_qubits[target].rb_fidelity = (
        results.fidelity[target],
        results.fit_uncertainties[target][1] / 2,
    )


standard_rb = Routine(_acquisition, _fit, _plot, _update)



"""CZ virtual correction experiment for two qubit gates, tune landscape."""

from dataclasses import dataclass, field
from typing import Literal, Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibolab import (
    AcquisitionType,
    AveragingMode,
    Delay,
    Parameter,
    Pulse,
    PulseSequence,
    Sweeper,
    VirtualZ,
)
from scipy.optimize import curve_fit

from qibocal.auto.operation import (
    Data,
    Parameters,
    QubitId,
    QubitPairId,
    Results,
    Routine,
)
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log
from qibocal.protocols.utils import table_dict, table_html

from ... import update
from ...update import replace
from .utils import order_pair


@dataclass
class VirtualZPhasesParameters(Parameters):
    """VirtualZ runcard inputs."""

    theta_start: float
    """Initial angle for the low frequency qubit measurement in radians."""
    theta_end: float
    """Final angle for the low frequency qubit measurement in radians."""
    theta_step: float
    """Step size for the theta sweep in radians."""
    native: str = "CZ"
    """Two qubit interaction to be calibrated.

    iSWAP and CZ are the possible options.

    """
    dt: Optional[float] = 0
    """Time delay between flux pulses and readout."""


@dataclass
class VirtualZPhasesResults(Results):
    """VirtualZ outputs when fitting will be done."""

    fitted_parameters: dict[tuple[str, QubitId],]
    """Fitted parameters"""
    native: str
    """Native two qubit gate."""
    angle: dict[QubitPairId, float]
    """Native angle."""
    virtual_phase: dict[QubitPairId, dict[QubitId, float]]
    """Virtual Z phase correction."""
    leakage: dict[QubitPairId, dict[QubitId, float]]
    """Leakage on control qubit for pair."""

    def __contains__(self, key: QubitPairId):
        """Check if key is in class.
        While key is a QubitPairId both chsh and chsh_mitigated contain
        an additional key which represents the basis chosen.
        """
        return key in [
            (target, control) for target, control, _ in self.fitted_parameters
        ]


VirtualZPhasesType = np.dtype([("target", np.float64), ("control", np.float64)])


@dataclass
class VirtualZPhasesData(Data):
    """VirtualZPhases data."""

    data: dict[tuple, npt.NDArray[VirtualZPhasesType]] = field(default_factory=dict)
    native: str = "CZ"
    thetas: list = field(default_factory=list)

    def __getitem__(self, pair):
        return {
            index: value
            for index, value in self.data.items()
            if set(pair).issubset(index)
        }


def create_sequence(
    platform: CalibrationPlatform,
    setup: Literal["I", "X"],
    target_qubit: QubitId,
    control_qubit: QubitId,
    ordered_pair: list[QubitId, QubitId],
    native: Literal["CZ", "iSWAP"],
    dt: float,
    flux_pulse_max_duration: float = None,
) -> tuple[
    PulseSequence,
    Pulse,
    Pulse,
]:
    """Create the experiment PulseSequence."""

    target_natives = platform.natives.single_qubit[target_qubit]
    control_natives = platform.natives.single_qubit[control_qubit]

    sequence = PulseSequence()
    # Y90
    sequence += target_natives.R(theta=np.pi / 2, phi=np.pi / 2)
    # X
    if setup == "X":
        sequence += control_natives.RX()
    else:
        sequence.append(
            (platform.qubits[control_qubit].drive, Delay(duration=sequence.duration))
        )

    drive_duration = sequence.duration

    # CZ
    flux_sequence = getattr(platform.natives.two_qubit[ordered_pair], native)()
    flux_pulses = [
        (ch, pulse) for ch, pulse in flux_sequence if not isinstance(pulse, VirtualZ)
    ]

    if flux_pulse_max_duration is not None:
        flux_pulses[0] = (
            flux_pulses[0][0],
            replace(flux_pulses[0][1], duration=flux_pulse_max_duration),
        )

    _, flux_pulse = flux_pulses[0]
    flux_sequence = PulseSequence(flux_pulses)
    sequence |= flux_sequence

    flux_duration = flux_sequence.duration
    dt_delay = Delay(duration=dt)

    theta_sequence = PulseSequence(
        [
            (
                platform.qubits[target_qubit].drive,
                dt_delay,
            ),
            (
                platform.qubits[control_qubit].drive,
                dt_delay,
            ),
            (
                platform.qubits[target_qubit].drive,
                Delay(duration=flux_duration),
            ),
            (
                platform.qubits[control_qubit].drive,
                Delay(duration=flux_duration),
            ),
        ]
    )
    # R90 (angle to be swept)
    theta_sequence += target_natives.R(theta=np.pi / 2, phi=0)
    theta_pulse = theta_sequence[-1][1]
    # X
    if setup == "X":
        theta_sequence += control_natives.RX()

    sequence += theta_sequence

    ro_target_delay = Delay(duration=flux_duration)
    ro_control_delay = Delay(duration=flux_duration)

    ro_sequence = PulseSequence(
        [
            (
                platform.qubits[target_qubit].acquisition,
                Delay(duration=drive_duration),
            ),
            (
                platform.qubits[control_qubit].acquisition,
                Delay(duration=drive_duration),
            ),
            (
                platform.qubits[target_qubit].acquisition,
                ro_target_delay,
            ),
            (
                platform.qubits[control_qubit].acquisition,
                ro_control_delay,
            ),
            (
                platform.qubits[target_qubit].acquisition,
                Delay(duration=theta_sequence.duration),
            ),
            (
                platform.qubits[control_qubit].acquisition,
                Delay(duration=theta_sequence.duration),
            ),
            target_natives.MZ()[0],
            control_natives.MZ()[0],
        ]
    )

    sequence += ro_sequence

    return sequence, flux_pulse, theta_pulse, [ro_target_delay, ro_control_delay]


def _acquisition(
    params: VirtualZPhasesParameters,
    platform: CalibrationPlatform,
    targets: list[QubitPairId],
) -> VirtualZPhasesData:
    r"""
    Acquisition for VirtualZPhases.

    Check the two-qubit landscape created by a flux pulse of a given duration
    and amplitude.
    The system is initialized with a Y90 pulse on the low frequency qubit and either
    an Id or an X gate on the high frequency qubit. Then the flux pulse is applied to
    the high frequency qubit in order to perform a two-qubit interaction. The Id/X gate
    is undone in the high frequency qubit and a theta90 pulse is applied to the low
    frequency qubit before measurement. That is, a pi-half pulse around the relative phase
    parametereized by the angle theta.
    Measurements on the low frequency qubit yield the 2Q-phase of the gate and the
    remnant single qubit Z phase aquired during the execution to be corrected.
    Population of the high frequency qubit yield the leakage to the non-computational states
    during the execution of the flux pulse.
    """

    theta_absolute = np.arange(params.theta_start, params.theta_end, params.theta_step)
    data = VirtualZPhasesData(thetas=theta_absolute.tolist(), native=params.native)
    for pair in targets:
        # order the qubits so that the low frequency one is the first
        ordered_pair = order_pair(pair, platform)

        for target_q, control_q in (
            (ordered_pair[0], ordered_pair[1]),
            (ordered_pair[1], ordered_pair[0]),
        ):
            for setup in ("I", "X"):
                (
                    sequence,
                    _,
                    theta_pulse,
                    _,
                ) = create_sequence(
                    platform,
                    setup,
                    target_q,
                    control_q,
                    ordered_pair,
                    params.native,
                    params.dt,
                )
                sweeper = Sweeper(
                    parameter=Parameter.relative_phase,
                    range=(params.theta_start, params.theta_end, params.theta_step),
                    pulses=[theta_pulse],
                )
                results = platform.execute(
                    [sequence],
                    [[sweeper]],
                    nshots=params.nshots,
                    relaxation_time=params.relaxation_time,
                    acquisition_type=AcquisitionType.DISCRIMINATION,
                    averaging_mode=AveragingMode.CYCLIC,
                )

                ro_target = list(
                    sequence.channel(platform.qubits[target_q].acquisition)
                )[-1]
                ro_control = list(
                    sequence.channel(platform.qubits[control_q].acquisition)
                )[-1]
                result_target = results[ro_target.id]
                result_control = results[ro_control.id]

                data.register_qubit(
                    VirtualZPhasesType,
                    (target_q, control_q, setup),
                    dict(
                        target=result_target,
                        control=result_control,
                    ),
                )
    return data


def sinusoid(x, amplitude, offset, phase):
    """Sinusoidal fit function."""
    return np.sin(x + phase) * amplitude + offset


def phase_diff(phase_1, phase_2):
    """Return the phase difference of two sinusoids, normalized in the range [0, pi]."""
    return np.arccos(np.cos(phase_1 - phase_2))


def fit_sinusoid(thetas, data):
    """Fit sinusoid to the given data."""
    pguess = [
        np.max(data) - np.min(data),
        np.mean(data),
        np.pi,
    ]

    popt, _ = curve_fit(
        sinusoid,
        thetas,
        data,
        p0=pguess,
        bounds=(
            (0, -np.max(data), 0),
            (np.max(data), np.max(data), 2 * np.pi),
        ),
    )

    return popt.tolist()


def _fit(
    data: VirtualZPhasesData,
) -> VirtualZPhasesResults:
    r"""Fitting routine for the experiment.

    The used model is

    .. math::

        y = p_0 sin\Big(x + p_2\Big) + p_1.
    """
    fitted_parameters = {}
    pairs = data.pairs
    virtual_phase = {}
    angle = {}
    leakage = {}
    for pair in pairs:
        virtual_phase[pair] = {}
        leakage[pair] = {}
        for target, control, setup in data[pair]:
            target_data = data[pair][target, control, setup].target
            try:
                params = fit_sinusoid(np.array(data.thetas), target_data)
                fitted_parameters[target, control, setup] = params

            except Exception as e:
                log.warning(f"CZ fit failed for pair ({target, control}) due to {e}.")

        try:
            for target_q, control_q in (
                pair,
                list(pair)[::-1],
            ):
                angle[target_q, control_q] = phase_diff(
                    fitted_parameters[target_q, control_q, "X"][2],
                    fitted_parameters[target_q, control_q, "I"][2],
                )
                virtual_phase[pair][target_q] = -fitted_parameters[
                    target_q, control_q, "I"
                ][2]

                # leakage estimate: L = m /2
                # See NZ paper from Di Carlo
                # approximation which does not need qutrits
                # https://arxiv.org/pdf/1903.02492.pdf
                leakage[pair][control_q] = 0.5 * float(
                    np.mean(
                        data[pair][target_q, control_q, "X"].control
                        - data[pair][target_q, control_q, "I"].control
                    )
                )
        except KeyError:
            pass  # exception covered above
    return VirtualZPhasesResults(
        native=data.native,
        angle=angle,
        virtual_phase=virtual_phase,
        fitted_parameters=fitted_parameters,
        leakage=leakage,
    )


def _plot(data: VirtualZPhasesData, fit: VirtualZPhasesResults, target: QubitPairId):
    """Plot routine for VirtualZPhases."""
    pair_data = data[target]
    qubits = next(iter(pair_data))[:2]
    fig1 = make_subplots(
        rows=1,
        cols=2,
        subplot_titles=(
            f"Qubit {qubits[0]}",
            f"Qubit {qubits[1]}",
        ),
    )
    fitting_report = set()
    fig2 = make_subplots(
        rows=1,
        cols=2,
        subplot_titles=(
            f"Qubit {qubits[0]}",
            f"Qubit {qubits[1]}",
        ),
    )

    thetas = data.thetas
    for target_q, control_q, setup in pair_data:
        target_prob = pair_data[target_q, control_q, setup].target
        control_prob = pair_data[target_q, control_q, setup].control
        fig = fig1 if (target_q, control_q) == qubits else fig2
        fig.add_trace(
            go.Scatter(
                x=np.array(thetas),
                y=target_prob,
                name=f"{setup} sequence",
                legendgroup=setup,
            ),
            row=1,
            col=1 if fig == fig1 else 2,
        )

        fig.add_trace(
            go.Scatter(
                x=np.array(thetas),
                y=control_prob,
                name=f"{setup} sequence",
                legendgroup=setup,
            ),
            row=1,
            col=2 if fig == fig1 else 1,
        )
        if fit is not None:
            angle_range = np.linspace(thetas[0], thetas[-1], 100)
            fitted_parameters = fit.fitted_parameters[target_q, control_q, setup]
            fig.add_trace(
                go.Scatter(
                    x=angle_range,
                    y=sinusoid(
                        angle_range,
                        *fitted_parameters,
                    ),
                    name="Fit",
                    line=go.scatter.Line(dash="dot"),
                ),
                row=1,
                col=1 if fig == fig1 else 2,
            )

            fitting_report.add(
                table_html(
                    table_dict(
                        [target_q, target_q, control_q],
                        [
                            f"{fit.native} angle [rad]",
                            "Virtual Z phase [rad]",
                            "Leakage [a.u.]",
                        ],
                        [
                            np.round(fit.angle[target_q, control_q], 4),
                            np.round(
                                fit.virtual_phase[tuple(sorted(target))][target_q], 4
                            ),
                            np.round(fit.leakage[tuple(sorted(target))][control_q], 4),
                        ],
                    )
                )
            )

    fig1.update_layout(
        title_text=f"Phase correction Qubit {qubits[0]}",
        showlegend=True,
        xaxis1_title="Virtual phase[rad]",
        xaxis2_title="Virtual phase [rad]",
        yaxis_title="State 1 Probability",
    )

    fig2.update_layout(
        title_text=f"Phase correction Qubit {qubits[1]}",
        showlegend=True,
        xaxis1_title="Virtual phase[rad]",
        xaxis2_title="Virtual phase[rad]",
        yaxis_title="State 1 Probability",
    )

    return [fig1, fig2], "".join(fitting_report)  # target and control qubit


def _update(
    results: VirtualZPhasesResults, platform: CalibrationPlatform, target: QubitPairId
):
    # FIXME: quick fix for qubit order
    target = tuple(sorted(target))
    update.virtual_phases(
        results.virtual_phase[target], results.native, platform, target
    )


correct_virtual_z_phases = Routine(
    _acquisition, _fit, _plot, _update, two_qubit_gates=True
)
"""Virtual phases correction protocol."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, PulseSequence
from scipy.signal import butter, filtfilt

from qibocal import update
from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform


@dataclass
class CalibrateStateDiscriminationParameters(Parameters):
    """Calibrate State Discrimination inputs."""

    """Frequency step for sweep (Hz)."""
    nshots: Optional[int] = None
    """Number of shots."""
    relaxation_time: Optional[int] = None
    """Relaxation time (ns)."""
    unrolling: Optional[bool] = False


CalibrateStateDiscriminationResType = np.dtype(
    [
        ("State 0 kernel", np.float64),
    ]
)
"""Custom dtype for CalibrateStateDiscrimination."""


def _get_lo_frequency(platform: CalibrationPlatform, qubit: QubitId) -> float:
    """Get LO frequency given QubitId.

    Currently it assumes that instruments with LOs is first one.
    """
    probe = platform.channels[platform.qubits[qubit].probe]
    lo_config = platform.config(probe.lo)
    return lo_config.frequency


@dataclass
class CalibrateStateDiscriminationResults(Results):
    """Calibrate State Discrimination outputs."""

    data: dict[
        tuple[QubitId, int], npt.NDArray[CalibrateStateDiscriminationResType]
    ] = field(default_factory=dict)
    """State 0 kernel"""


CalibrateStateDiscriminationType = np.dtype(
    [
        ("i", np.float64),
        ("q", np.float64),
    ]
)
"""Custom dtype for CalibrateStateDiscrimination."""


@dataclass
class CalibrateStateDiscriminationData(Data):
    """CalibrateStateDiscrimination acquisition outputs."""

    resonator_type: str
    """Resonator type."""
    intermediate_frequency: dict[QubitId, float] = field(default_factory=dict)
    """Intermediate RO frequency for each qubit."""
    data: dict[tuple[QubitId, int], npt.NDArray[CalibrateStateDiscriminationType]] = (
        field(default_factory=dict)
    )


def _acquisition(
    params: CalibrateStateDiscriminationParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> CalibrateStateDiscriminationData:
    r"""
    Data acquisition for Calibrate State Discrimination experiment.
    Calculates the optimal kernel for the readout. It has to be run one qubit at a time.
    The kernels are stored in the result.npz generated on the report.

    Args:
        params (CalibrateStateDiscriminationParameters): experiment's parameters
        platform (CalibrationPlatform): Qibolab platform object
        qubits (dict): list of target qubits to perform the action

    """

    native = platform.natives.single_qubit
    sequences, all_ro_pulses = [], []
    for state in [0, 1]:
        ro_pulses = {}
        sequence = PulseSequence()
        for q in targets:
            ro_sequence = native[q].MZ()
            ro_pulses[q] = ro_sequence[0][1].id
            sequence += ro_sequence

        if state == 1:
            rx_sequence = PulseSequence()
            for q in targets:
                rx_sequence += native[q].RX()
            sequence = rx_sequence | sequence

        sequences.append(sequence)
        all_ro_pulses.append(ro_pulses)

    options = dict(
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.RAW,
        averaging_mode=AveragingMode.CYCLIC,
    )

    if params.unrolling:
        results = platform.execute(sequences, **options)
    else:
        results = {}
        for sequence in sequences:
            results.update(platform.execute([sequence], **options))

    intermediate_frequency = {
        qubit: platform.config(platform.qubits[qubit].probe).frequency
        - _get_lo_frequency(platform, qubit)
        for qubit in targets
    }
    data = CalibrateStateDiscriminationData(
        resonator_type=platform.resonator_type,
        intermediate_frequency=intermediate_frequency,
    )
    for state, ro_pulses in zip([0, 1], all_ro_pulses):
        for qubit in targets:
            serial = ro_pulses[qubit]
            result = results[serial]
            data.register_qubit(
                CalibrateStateDiscriminationType,
                (qubit, state),
                dict(
                    i=result[..., 0],
                    q=result[..., 1],
                ),
            )
    return data


def _fit(data: CalibrateStateDiscriminationData) -> CalibrateStateDiscriminationResults:
    """Post-Processing for Calibrate State Discrimination.

    The raw traces for 0 and 1 are demodulated with the IF stored during acquisition.
    Moreover, fast oscillating terms (:math:`\\exp(-2\\pi \\omega_{\text{IF} t)`) are removed by applying a lowpass filter.
    Finally, the optimal kernel is calculated as the difference between the two traces.
    """
    qubits = data.qubits

    kernel_state_zero = {}

    def lowpass_filter(signal, cutoff, fs=1e9, order=5):
        """Basic lowpass filter."""
        norm_cutoff = cutoff / fs
        b, a = butter(order, norm_cutoff)
        return filtfilt(b, a, signal)

    for qubit in qubits:
        traces = []
        freq = data.intermediate_frequency[qubit]

        for i in range(2):
            trace = data[qubit, i].i + 1.0j * data[qubit, i].q
            t = np.arange(0, len(trace), 1)

            # demodulation (we assume that RAW doesn't demodulate)
            trace = np.array(
                [
                    np.exp(-2 * np.pi * t[i] * 1.0j * freq * 1e-9) * trace[i]
                    for i in range(len(t))
                ]
            )
            # apply lowpass filter to remove fast rotating terms
            trace = lowpass_filter(trace, 2 * np.abs(freq))
            traces.append(trace)

        # Calculate the optimal kernel
        kernel = np.conj(traces[0] - traces[1])

        # Normalize the kernel
        norm = np.sqrt(np.sum(np.abs(kernel) ** 2))
        max_abs_weight = np.max(np.abs(kernel / norm))
        kernel /= norm * max_abs_weight
        kernel_state_zero[qubit] = kernel

    return CalibrateStateDiscriminationResults(data=kernel_state_zero)


def _plot(
    data: CalibrateStateDiscriminationData,
    target,
    fit: CalibrateStateDiscriminationResults,
):
    """Plotting function for Calibrate State Discrimination."""
    # Plot kernels
    figures = []
    fitting_report = ""
    trace0 = data[target, 0].i + 1.0j * data[target, 0].q
    t = np.arange(0, len(trace0), 1)

    if fit is not None:
        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=t,
                y=fit.data[target].real,
                opacity=1,
                name="Real",
                showlegend=True,
            )
        )

        fig.add_trace(
            go.Scatter(
                x=t,
                y=fit.data[target].imag,
                opacity=1,
                name="Imag",
                showlegend=True,
            )
        )

        fig.update_layout(
            showlegend=True,
            title="Optimal integration kernel",
            xaxis_title="Readout sample",
            yaxis_title="Normalized weight",
        )

        figures.append(fig)

    return figures, fitting_report


def _update(
    results: CalibrateStateDiscriminationResults,
    platform: CalibrationPlatform,
    qubit: QubitId,
):
    update.kernel(results.data[qubit], platform, qubit)


calibrate_state_discrimination = Routine(_acquisition, _fit, _plot, _update)
"""Calibrate State Discrimination Routine object."""



"""Specify runcard layout, handles (de)serialization."""

from dataclasses import asdict
from pathlib import Path
from typing import Any, Optional, Union

import yaml
from pydantic.dataclasses import dataclass
from qibolab import Platform

from .. import protocols
from .execute import Executor
from .history import History
from .mode import ExecutionMode
from .task import Action, Targets

RUNCARD = "runcard.yml"
"""Runcard filename."""


@dataclass
class Runcard:
    """Structure of an execution runcard."""

    actions: list[Action]
    """List of action to be executed."""
    targets: Optional[Targets] = None
    """Qubits to be calibrated.

    If `None` the protocols will be executed on all qubits
    available in the platform.
    """
    backend: str = "qibolab"
    """Qibo backend."""
    platform: str = "mock"
    """Qibolab platform."""
    update: bool = True

    @classmethod
    def load(cls, runcard: Union[dict[str, Any], Path]):
        """Load a runcard dict or path."""
        if not isinstance(runcard, dict):
            return cls(yaml.safe_load((runcard / RUNCARD).read_text(encoding="utf-8")))
        return cls(**runcard)

    def dump(self, path):
        """Dump runcard object to yaml."""
        (path / RUNCARD).write_text(yaml.safe_dump(asdict(self)))

    def run(
        self, output: Path, platform: Platform, mode: ExecutionMode, update: bool = True
    ) -> History:
        """Run runcard and dump to output."""
        targets = self.targets if self.targets is not None else list(platform.qubits)
        history = History.load(output)
        update = update and self.update
        instance = Executor(
            history=history, platform=platform, targets=targets, update=update
        )

        for action in self.actions:
            instance.run_protocol(
                protocol=getattr(protocols, action.operation),
                parameters=action,
                mode=mode,
            )
            instance.history.flush(output)
        return instance.history



from dataclasses import dataclass, field
from typing import Optional, Union

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Parameter, Readout, Sweeper

from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log
from qibocal.result import magnitude

from ... import update
from ..utils import readout_frequency, table_dict, table_html
from .utils import fitting, process_fit, ramsey_fit, ramsey_sequence


@dataclass
class RamseySignalParameters(Parameters):
    """Ramsey runcard inputs."""

    delay_between_pulses_start: int
    """Initial delay between RX(pi/2) pulses in ns."""
    delay_between_pulses_end: int
    """Final delay between RX(pi/2) pulses in ns."""
    delay_between_pulses_step: int
    """Step delay between RX(pi/2) pulses in ns."""
    detuning: Optional[int] = None
    """Frequency detuning [Hz] (optional).
        If 0 standard Ramsey experiment is performed."""
    unrolling: bool = False
    """If ``True`` it uses sequence unrolling to deploy multiple sequences in a single instrument call.
    Defaults to ``False``."""


@dataclass
class RamseySignalResults(Results):
    """Ramsey outputs."""

    detuning: float
    """Qubit frequency detuning."""
    frequency: dict[QubitId, Union[float, list[float]]]
    """Drive frequency [GHz] for each qubit."""
    t2: dict[QubitId, Union[float, list[float]]]
    """T2 for each qubit [ns]."""
    delta_phys: dict[QubitId, Union[float, list[float]]]
    """Drive frequency [Hz] correction for each qubit."""
    delta_fitting: dict[QubitId, Union[float, list[float]]]
    """Raw drive frequency [Hz] correction for each qubit.
       including the detuning."""
    fitted_parameters: dict[QubitId, list[float]]
    """Raw fitting output."""


RamseySignalType = np.dtype([("wait", np.float64), ("signal", np.float64)])
"""Custom dtype for coherence routines."""


@dataclass
class RamseySignalData(Data):
    """Ramsey acquisition outputs."""

    detuning: Optional[int] = None
    """Frequency detuning [Hz]."""
    qubit_freqs: dict[QubitId, float] = field(default_factory=dict)
    """Qubit freqs for each qubit."""
    data: dict[QubitId, npt.NDArray[RamseySignalType]] = field(default_factory=dict)
    """Raw data acquired."""

    @property
    def waits(self):
        """
        Return a list with the waiting times without repetitions.
        """
        qubit = next(iter(self.data))
        return np.unique(self.data[qubit].wait)


def _acquisition(
    params: RamseySignalParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> RamseySignalData:
    """Data acquisition for Ramsey Experiment (detuned)."""
    # create a sequence of pulses for the experiment
    # RX90 - t - RX90 - MZ
    # define the parameter to sweep and its range:

    waits = np.arange(
        # wait time between RX90 pulses
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    data = RamseySignalData(
        detuning=params.detuning,
        qubit_freqs={
            qubit: platform.config(platform.qubits[qubit].drive).frequency
            for qubit in targets
        },
    )

    updates = []
    updates += [
        {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
        for q in targets
    ]
    if params.detuning is not None:
        for qubit in targets:
            channel = platform.qubits[qubit].drive
            f0 = platform.config(channel).frequency
            updates.append({channel: {"frequency": f0 + params.detuning}})

    if not params.unrolling:
        sequence, delays = ramsey_sequence(platform, targets)
        sweeper = Sweeper(
            parameter=Parameter.duration,
            values=waits,
            pulses=delays,
        )

        # execute the sweep
        results = platform.execute(
            [sequence],
            [[sweeper]],
            updates=updates,
            relaxation_time=params.relaxation_time,
            acquisition_type=AcquisitionType.INTEGRATION,
            averaging_mode=AveragingMode.CYCLIC,
        )
        for qubit in targets:
            ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[-1]
            result = results[ro_pulse.id]
            # The probability errors are the standard errors of the binomial distribution
            data.register_qubit(
                RamseySignalType,
                (qubit),
                dict(
                    wait=waits,
                    signal=magnitude(result),
                ),
            )

    else:
        sequences, all_ro_pulses = [], []
        for wait in waits:
            sequence, _ = ramsey_sequence(platform, targets, wait)
            sequences.append(sequence)
            all_ro_pulses.append(
                {
                    qubit: [
                        pulse
                        for pulse in list(
                            sequence.channel(platform.qubits[qubit].acquisition)
                        )
                        if isinstance(pulse, Readout)
                    ][0]
                    for qubit in targets
                }
            )

        results = platform.execute(
            sequences,
            nshots=params.nshots,
            relaxation_time=params.relaxation_time,
            acquisition_type=AcquisitionType.INTEGRATION,
            averaging_mode=AveragingMode.CYCLIC,
            updates=updates,
        )

        for wait, ro_pulses in zip(waits, all_ro_pulses):
            for qubit in targets:
                result = results[ro_pulses[qubit].id]
                data.register_qubit(
                    RamseySignalType,
                    (qubit),
                    dict(
                        wait=np.array([wait]),
                        signal=np.array([magnitude(result)]),
                    ),
                )

    return data


def _fit(data: RamseySignalData) -> RamseySignalResults:
    r"""Fitting routine for Ramsey experiment. The used model is
    .. math::

        y = p_0 + p_1 sin \Big(p_2 x + p_3 \Big) e^{-x p_4}.
    """
    qubits = data.qubits
    waits = data.waits
    popts = {}
    freq_measure = {}
    t2_measure = {}
    delta_phys_measure = {}
    delta_fitting_measure = {}
    for qubit in qubits:
        qubit_data = data[qubit]
        qubit_freq = data.qubit_freqs[qubit]
        signal = qubit_data["signal"]
        try:
            popt, perr = fitting(waits, signal)
            (
                freq_measure[qubit],
                t2_measure[qubit],
                delta_phys_measure[qubit],
                delta_fitting_measure[qubit],
                popts[qubit],
            ) = process_fit(popt, perr, qubit_freq, data.detuning)
        except Exception as e:
            log.warning(f"Ramsey fitting failed for qubit {qubit} due to {e}.")

    return RamseySignalResults(
        detuning=data.detuning,
        frequency=freq_measure,
        t2=t2_measure,
        delta_phys=delta_phys_measure,
        delta_fitting=delta_fitting_measure,
        fitted_parameters=popts,
    )


def _plot(data: RamseySignalData, target: QubitId, fit: RamseySignalResults = None):
    """Plotting function for Ramsey Experiment."""

    figures = []
    fig = go.Figure()
    fitting_report = ""

    qubit_data = data.data[target]
    waits = data.waits
    signal = qubit_data["signal"]
    fig = go.Figure(
        [
            go.Scatter(
                x=waits,
                y=signal,
                opacity=1,
                name="Signal",
                showlegend=True,
                legendgroup="Signal",
                mode="lines",
            ),
        ]
    )

    if fit is not None:
        fig.add_trace(
            go.Scatter(
                x=waits,
                y=ramsey_fit(
                    waits,
                    float(fit.fitted_parameters[target][0]),
                    float(fit.fitted_parameters[target][1]),
                    float(fit.fitted_parameters[target][2]),
                    float(fit.fitted_parameters[target][3]),
                    float(fit.fitted_parameters[target][4]),
                ),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                target,
                [
                    "Delta Frequency [Hz]",
                    "Delta Frequency (with detuning) [Hz]",
                    "Drive Frequency [Hz]",
                    "T2* [ns]",
                ],
                [
                    np.round(fit.delta_phys[target][0], 3),
                    np.round(fit.delta_fitting[target][0], 3),
                    np.round(fit.frequency[target][0], 3),
                    np.round(fit.t2[target][0], 3),
                ],
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Signal [a.u.]",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(
    results: RamseySignalResults, platform: CalibrationPlatform, target: QubitId
):
    if results.detuning is not None:
        update.drive_frequency(results.frequency[target][0], platform, target)
    else:
        update.t2(results.t2[target], platform, target)


ramsey_signal = Routine(_acquisition, _fit, _plot, _update)
"""Ramsey Routine object."""



from dataclasses import dataclass, field
from typing import Optional

from qibolab import AcquisitionType, PulseSequence

from qibocal.auto.operation import QubitId, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.protocols.classification import (
    ClassificationType,
    SingleShotClassificationData,
    SingleShotClassificationParameters,
)
from qibocal.protocols.utils import plot_results, readout_frequency

from ..auto.operation import Results

COLUMNWIDTH = 600
LEGEND_FONT_SIZE = 20
TITLE_SIZE = 25
SPACING = 0.1
DEFAULT_CLASSIFIER = "naive_bayes"


@dataclass
class QutritClassificationParameters(SingleShotClassificationParameters):
    """SingleShotClassification runcard inputs."""


@dataclass
class QutritClassificationData(SingleShotClassificationData):
    classifiers_list: Optional[list[str]] = field(
        default_factory=lambda: [DEFAULT_CLASSIFIER]
    )
    """List of models to classify the qubit states"""


@dataclass
class QutritClassificationResults(Results):
    """Qutrit classification results"""


def _acquisition(
    params: QutritClassificationParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> QutritClassificationData:
    """
    This Routine prepares the qubits in 0,1 and 2 states and measures their
    respective I, Q values.

    Args:
        nshots (int): number of times the pulse sequence will be repeated.
        classifiers (list): list of classifiers, the available ones are:
            - naive_bayes
            - nn
            - random_forest
            - decision_tree
        The default value is `["naive_bayes"]`.
        savedir (str): Dumping folder of the classification results.
        If not given, the dumping folder will be the report one.
        relaxation_time (float): Relaxation time.
    """

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    states = [0, 1, 2]
    sequences, all_ro_pulses = [], []
    native = platform.natives.single_qubit

    for state in states:
        ro_pulses = {}
        sequence = PulseSequence()
        for q in targets:
            ro_sequence = native[q].MZ()
            ro_pulses[q] = ro_sequence[0][1].id
            sequence += ro_sequence

        if state == 1:
            rx_sequence = PulseSequence()
            for q in targets:
                rx_sequence += native[q].RX()
            sequence = rx_sequence | sequence

        if state == 2:
            rx12_sequence = PulseSequence()
            for q in targets:
                rx12_sequence += native[q].RX() | native[q].RX12()
            sequence = rx12_sequence | sequence

        sequences.append(sequence)
        all_ro_pulses.append(ro_pulses)

    data = QutritClassificationData(
        nshots=params.nshots,
        classifiers_list=params.classifiers_list,
        savedir=params.savedir,
    )

    options = dict(
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
    )
    updates = [
        {
            platform.qubits[q].probe: {
                "frequency": readout_frequency(q, platform, state=1)
            }
        }
        for q in targets
    ]
    if params.unrolling:
        results = platform.execute(sequences, **options, updates=updates)
    else:
        results = {}
        for sequence in sequences:
            results.update(platform.execute([sequence], **options, updates=updates))

    for state, ro_pulses in zip(states, all_ro_pulses):
        for qubit in targets:
            serial = ro_pulses[qubit]
            result = results[serial]
            data.register_qubit(
                ClassificationType,
                (qubit),
                dict(
                    i=result[..., 0],
                    q=result[..., 1],
                    state=[state] * params.nshots,
                ),
            )

    return data


def _fit(data: QutritClassificationData) -> QutritClassificationResults:
    return QutritClassificationResults()


def _plot(
    data: QutritClassificationData,
    target: QubitId,
    fit: QutritClassificationResults,
):
    figures = plot_results(data, target, 3, None)
    fitting_report = ""
    return figures, fitting_report


qutrit_classification = Routine(_acquisition, _fit, _plot)
"""Qutrit classification Routine object."""



from .chevron import chevron as chevron
from .chevron import chevron_signal as chevron_signal
from .cryoscope import cryoscope as cryoscope
from .optimize import optimize_two_qubit_gate as optimize_two_qubit_gate
from .virtual_z_phases import correct_virtual_z_phases as correct_virtual_z_phases



from dataclasses import dataclass
from typing import Union

import numpy as np
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal.auto.operation import Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.result import magnitude, phase

from ... import update
from ..utils import readout_frequency, table_dict, table_html
from .t1_signal import T1SignalData
from .utils import (
    CoherenceType,
    dynamical_decoupling_sequence,
    exp_decay,
    exponential_fit,
)


@dataclass
class SpinEchoSignalParameters(Parameters):
    """SpinEcho Signal runcard inputs."""

    delay_between_pulses_start: int
    """Initial delay between pulses [ns]."""
    delay_between_pulses_end: int
    """Final delay between pulses [ns]."""
    delay_between_pulses_step: int
    """Step delay between pulses [ns]."""
    single_shot: bool = False


@dataclass
class SpinEchoSignalResults(Results):
    """SpinEchoSignal outputs."""

    t2: dict[QubitId, Union[float, list[float]]]
    """T2 echo for each qubit."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""
    pcov: dict[QubitId, list[float]]
    """Approximate covariance of fitted parameters."""


class SpinEchoSignalData(T1SignalData):
    """SpinEcho acquisition outputs."""


def _acquisition(
    params: SpinEchoSignalParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> SpinEchoSignalData:
    """Data acquisition for SpinEcho"""
    # create a sequence of pulses for the experiment:
    sequence, delays = dynamical_decoupling_sequence(platform, targets, kind="CP")

    # define the parameter to sweep and its range:
    # delay between pulses
    wait_range = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    durations = []
    for q in targets:
        # this is assuming that RX and RX90 have the same duration
        duration = platform.natives.single_qubit[q].RX()[0][1].duration
        durations.append(duration)
        assert (params.delay_between_pulses_start - duration) / 2 >= 0, (
            f"Initial delay too short for qubit {q}, minimum delay should be {duration}"
        )

    assert len(set(durations)) == 1, (
        "Cannot run on mulitple qubit with different RX duration."
    )

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=(wait_range - durations[0]) / 2,
        pulses=delays,
    )

    results = platform.execute(
        [sequence],
        [[sweeper]],
        updates=[
            {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
            for q in targets
        ],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=(
            AveragingMode.SINGLESHOT if params.single_shot else AveragingMode.CYCLIC
        ),
    )

    data = SpinEchoSignalData()
    for qubit in targets:
        ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[-1]
        result = results[ro_pulse.id]
        signal = magnitude(result)
        if params.single_shot:
            _wait = np.array(len(signal) * [wait_range])
        else:
            _wait = wait_range
        data.register_qubit(
            CoherenceType,
            (qubit),
            dict(
                wait=_wait,
                signal=signal,
                phase=phase(result),
            ),
        )

    return data


def _fit(data: SpinEchoSignalData) -> SpinEchoSignalResults:
    """Post-processing for SpinEcho."""
    data = data.average

    t2echos, fitted_parameters, pcov = exponential_fit(data)

    return SpinEchoSignalResults(t2echos, fitted_parameters, pcov)


def _plot(data: SpinEchoSignalData, target: QubitId, fit: SpinEchoSignalResults = None):
    """Plotting for SpinEcho"""
    data = data.average

    figures = []
    fig = go.Figure()

    # iterate over multiple data folders
    fitting_report = None

    qubit_data = data[target]
    waits = qubit_data.wait

    fig.add_trace(
        go.Scatter(
            x=waits,
            y=qubit_data.signal,
            opacity=1,
            name="Signal",
            showlegend=True,
            legendgroup="Signal",
        ),
    )

    if fit is not None:
        # add fitting trace
        waitrange = np.linspace(
            min(waits),
            max(waits),
            2 * len(qubit_data),
        )
        params = fit.fitted_parameters[target]

        fig.add_trace(
            go.Scatter(
                x=waitrange,
                y=exp_decay(waitrange, *params),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            ),
        )

        fitting_report = table_html(
            table_dict(
                target,
                ["T2 Spin Echo [ns]"],
                [np.round(fit.t2[target])],
                display_error=True,
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Signal [a.u.]",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(
    results: SpinEchoSignalResults, platform: CalibrationPlatform, target: QubitId
):
    update.t2_spin_echo(results.t2[target], platform, target)


spin_echo_signal = Routine(_acquisition, _fit, _plot, _update)
"""SpinEcho Routine object."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, PulseSequence, Readout

from qibocal import update
from qibocal.auto.operation import Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform

from ...result import probability
from ..utils import COLORBAND, COLORBAND_LINE, table_dict, table_html
from . import t1, utils


@dataclass
class ZenoParameters(Parameters):
    """Zeno runcard inputs."""

    readouts: int
    "Number of readout pulses"


@dataclass
class ZenoResults(Results):
    """Zeno outputs."""

    zeno_t1: dict[QubitId, int]
    """T1 for each qubit."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""
    pcov: dict[QubitId, list[float]]
    """Approximate covariance of fitted parameters."""
    chi2: dict[QubitId, tuple[float, Optional[float]]]
    """Chi squared estimate mean value and error."""


def zeno_sequence(
    platform: CalibrationPlatform, targets: list[QubitId], readouts: int
) -> tuple[PulseSequence, dict[QubitId, int]]:
    """Generating sequence for Zeno experiment."""

    sequence = PulseSequence()
    readout_duration = {}
    for q in targets:
        natives = platform.natives.single_qubit[q]
        _, ro_pulse = natives.MZ()[0]
        readout_duration[q] = ro_pulse.duration
        qubit_sequence = natives.RX() | natives.MZ()
        for _ in range(readouts - 1):
            qubit_sequence += natives.MZ()
        sequence += qubit_sequence

    return sequence, readout_duration


@dataclass
class ZenoData(t1.T1Data):
    readout_duration: dict[QubitId, float] = field(default_factory=dict)
    """Readout durations for each qubit"""


def _acquisition(
    params: ZenoParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> ZenoData:
    """
    In a T1_Zeno experiment, we measure an excited qubit repeatedly. Due to decoherence processes,
    it is possible that, at the time of measurement, the qubit will not be excited anymore.
    The quantum zeno effect consists of measuring allowing a particle's time evolution to be slowed
    down by measuring it frequently enough. However, in the experiments we see that due the QND-ness of the readout
    pulse that the qubit decoheres faster.
    Reference: https://link.aps.org/accepted/10.1103/PhysRevLett.118.240401.
    """

    sequence, readout_duration = zeno_sequence(
        platform, targets, readouts=params.readouts
    )
    data = ZenoData(readout_duration=readout_duration)

    results = platform.execute(
        [sequence],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.DISCRIMINATION,
        averaging_mode=AveragingMode.SINGLESHOT,
    )

    for qubit in targets:
        probs = []
        readouts = [
            pulse
            for pulse in sequence.channel(platform.qubits[qubit].acquisition)
            if isinstance(pulse, Readout)
        ]
        for i in range(params.readouts):
            ro_pulse = readouts[i]
            probs.append(probability(results[ro_pulse.id], state=1))

        data.register_qubit(
            t1.CoherenceProbType,
            (qubit),
            dict(
                wait=np.arange(params.readouts) + 1,
                prob=np.array(probs),
                error=np.array(
                    [np.sqrt(prob * (1 - prob) / params.nshots) for prob in probs]
                ),
            ),
        )
    return data


def _fit(data: ZenoData) -> ZenoResults:
    """
    Fitting routine for T1 experiment. The used model is

        .. math::

            y = p_0-p_1 e^{-x p_2}.
    """
    t1s, fitted_parameters, pcovs, chi2 = utils.exponential_fit_probability(data)
    return ZenoResults(t1s, fitted_parameters, pcovs, chi2)


def _plot(data: ZenoData, fit: ZenoResults, target: QubitId):
    """Plotting function for T1 experiment."""

    figures = []
    fitting_report = ""
    qubit_data = data[target]
    probs = qubit_data.prob
    error_bars = qubit_data.error
    readouts = qubit_data.wait

    fig = go.Figure(
        [
            go.Scatter(
                x=readouts,
                y=probs,
                opacity=1,
                name="Probability of 1",
                showlegend=True,
                legendgroup="Probability of 1",
                mode="lines",
            ),
            go.Scatter(
                x=np.concatenate((readouts, readouts[::-1])),
                y=np.concatenate((probs + error_bars, (probs - error_bars)[::-1])),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                showlegend=True,
                name="Errors",
            ),
        ]
    )

    if fit is not None:
        fitting_report = ""
        waitrange = np.linspace(
            min(readouts),
            max(readouts),
            2 * len(qubit_data),
        )
        params = fit.fitted_parameters[target]
        fig.add_trace(
            go.Scatter(
                x=waitrange,
                y=utils.exp_decay(waitrange, *params),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                target,
                ["T1 [ns]", "Readout Pulse [ns]", "chi2 reduced"],
                [
                    np.array(fit.zeno_t1[target]) * data.readout_duration[target],
                    (data.readout_duration[target], 0),
                    fit.chi2[target],
                ],
                display_error=True,
            )
        )
        # FIXME: Pulse duration (+ time of flight ?)

    # last part
    fig.update_layout(
        showlegend=True,
        xaxis_title="Number of readouts",
        yaxis_title="Probability of State 1",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(results: ZenoResults, platform: CalibrationPlatform, qubit: QubitId):
    update.t1(results.zeno_t1[qubit], platform, qubit)


zeno = Routine(_acquisition, _fit, _plot, _update)



from dataclasses import dataclass
from typing import Optional

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibolab import (
    AcquisitionType,
    AveragingMode,
    Delay,
    Parameter,
    PulseSequence,
    Sweeper,
)

from qibocal.auto.operation import Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform

from ..result import magnitude, phase
from ..update import replace
from .qubit_spectroscopy import QubitSpectroscopyResults
from .resonator_punchout import ResonatorPunchoutData
from .utils import HZ_TO_GHZ, readout_frequency


@dataclass
class QubitPowerSpectroscopyParameters(Parameters):
    """QubitPowerSpectroscopy runcard inputs."""

    freq_width: int
    """Width for frequency sweep relative  to the drive frequency [Hz]."""
    freq_step: int
    """Frequency step for sweep [Hz]."""
    min_amp: float
    """Minimum amplitude."""
    max_amp: float
    """Maximum amplitude."""
    step_amp: float
    """Step amplitude."""
    duration: int
    """Drive duration."""


@dataclass
class QubitPowerSpectroscopyData(ResonatorPunchoutData):
    """QubitPowerSpectroscopy data acquisition."""


def _acquisition(
    params: QubitPowerSpectroscopyParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> QubitPowerSpectroscopyData:
    """Perform a qubit spectroscopy experiment with different amplitudes.

    For high amplitude it should be possible to see more peaks: corresponding to
    the (0-2)/2 frequency and the 1-2.
    This experiment can be used also to test if a peak is a qubit: if it is, the
    peak will get larger while increasing the power of the drive.
    """
    # define the sequence: RX - MZ
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    freq_sweepers = {}
    delta_frequency_range = np.arange(
        -params.freq_width / 2, params.freq_width / 2, params.freq_step
    )
    for qubit in targets:
        natives = platform.natives.single_qubit[qubit]
        qd_channel, qd_pulse = natives.RX()[0]
        ro_channel, ro_pulse = natives.MZ()[0]

        qd_pulse = replace(qd_pulse, duration=params.duration)

        qd_pulses[qubit] = qd_pulse
        ro_pulses[qubit] = ro_pulse

        sequence.append((qd_channel, qd_pulse))
        sequence.append((ro_channel, Delay(duration=qd_pulse.duration)))
        sequence.append((ro_channel, ro_pulse))

        f0 = platform.config(qd_channel).frequency
        freq_sweepers[qubit] = Sweeper(
            parameter=Parameter.frequency,
            values=f0 + delta_frequency_range,
            channels=[qd_channel],
        )

    amp_sweeper = Sweeper(
        parameter=Parameter.amplitude,
        range=(params.min_amp, params.max_amp, params.step_amp),
        pulses=[qd_pulses[qubit] for qubit in targets],
    )

    # data
    data = QubitPowerSpectroscopyData(
        resonator_type=platform.resonator_type,
    )

    results = platform.execute(
        [sequence],
        [[amp_sweeper], [freq_sweepers[q] for q in targets]],
        updates=[
            {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
            for q in targets
        ],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=AveragingMode.CYCLIC,
    )

    # retrieve the results for every qubit
    for qubit, ro_pulse in ro_pulses.items():
        # average signal, phase, i and q over the number of shots defined in the runcard
        result = results[ro_pulse.id]
        data.register_qubit(
            qubit,
            signal=magnitude(result),
            phase=phase(result),
            freq=freq_sweepers[qubit].values,
            amp=amp_sweeper.values,
        )

    return data


def _fit(data: QubitPowerSpectroscopyData) -> Results:
    """Do not perform any fitting procedure."""
    return Results()


def _plot(
    data: ResonatorPunchoutData,
    target: QubitId,
    fit: Optional[QubitSpectroscopyResults] = None,
):
    """Plot QubitPunchout."""
    figures = []
    fitting_report = ""
    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
        subplot_titles=(
            "Signal [a.u.]",
            "phase [rad]",
        ),
    )
    qubit_data = data[target]
    frequencies = qubit_data.freq * HZ_TO_GHZ
    amplitudes = qubit_data.amp

    fig.add_trace(
        go.Heatmap(
            x=frequencies,
            y=amplitudes,
            z=qubit_data.signal,
            colorbar_x=0.46,
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Heatmap(
            x=frequencies,
            y=amplitudes,
            z=qubit_data.phase,
            colorbar_x=1.01,
        ),
        row=1,
        col=2,
    )

    fig.update_layout(
        showlegend=True,
        legend=dict(orientation="h"),
    )

    fig.update_xaxes(title_text="Drive frequency [GHz]", row=1, col=1)
    fig.update_xaxes(title_text="Drive frequency [GHz]", row=1, col=2)
    fig.update_yaxes(title_text="Drive amplitude [a.u.]", row=1, col=1)

    figures.append(fig)

    return figures, fitting_report


qubit_power_spectroscopy = Routine(_acquisition, _fit, _plot)
"""QubitPowerSpectroscopy Routine object."""



from dataclasses import dataclass, field
from typing import Optional

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Delay, Drag, PulseSequence
from scipy.optimize import curve_fit

from qibocal import update
from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.config import log
from qibocal.result import probability
from qibocal.update import replace

from .utils import (
    COLORBAND,
    COLORBAND_LINE,
    chi2_reduced,
    fallback_period,
    guess_period,
    table_dict,
    table_html,
)


# TODO: add errors in fitting
@dataclass
class DragTuningParameters(Parameters):
    """DragTuning runcard inputs."""

    beta_start: float
    """DRAG pulse beta start sweep parameter."""
    beta_end: float
    """DRAG pulse beta end sweep parameter."""
    beta_step: float
    """DRAG pulse beta sweep step parameter."""
    unrolling: bool = False
    """If ``True`` it uses sequence unrolling to deploy multiple sequences in a single instrument call.
    Defaults to ``False``."""
    nflips: int = 1
    """Repetitions of (Xpi - Xmpi)."""


@dataclass
class DragTuningResults(Results):
    """DragTuning outputs."""

    betas: dict[QubitId, float]
    """Optimal beta paramter for each qubit."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""
    chi2: dict[QubitId, tuple[float, Optional[float]]] = field(default_factory=dict)
    """Chi2 calculation."""


DragTuningType = np.dtype(
    [("prob", np.float64), ("error", np.float64), ("beta", np.float64)]
)


@dataclass
class DragTuningData(Data):
    """DragTuning acquisition outputs."""

    data: dict[QubitId, npt.NDArray[DragTuningType]] = field(default_factory=dict)
    """Raw data acquired."""


def _acquisition(
    params: DragTuningParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> DragTuningData:
    r"""
    Data acquisition for drag pulse tuning experiment.
    See https://arxiv.org/pdf/1504.06597.pdf Fig. 2 (c).
    """

    data = DragTuningData()
    beta_param_range = np.arange(params.beta_start, params.beta_end, params.beta_step)

    sequences, all_ro_pulses = [], []
    for beta_param in beta_param_range:
        sequence = PulseSequence()
        ro_pulses = {}
        for q in targets:
            natives = platform.natives.single_qubit[q]
            qd_channel, qd_pulse = natives.RX()[0]
            ro_channel, ro_pulse = natives.MZ()[0]

            drag = replace(
                qd_pulse,
                envelope=Drag(
                    rel_sigma=qd_pulse.envelope.rel_sigma,
                    beta=beta_param,
                ),
            )
            drag_negative = replace(drag, relative_phase=np.pi)

            for _ in range(params.nflips):
                sequence.append((qd_channel, drag))
                sequence.append((qd_channel, drag_negative))
            sequence.append(
                (
                    ro_channel,
                    Delay(
                        duration=params.nflips
                        * (drag.duration + drag_negative.duration)
                    ),
                )
            )
            sequence.append((ro_channel, ro_pulse))

        sequences.append(sequence)
        all_ro_pulses.append(
            {
                qubit: list(sequence.channel(platform.qubits[q].acquisition))[-1]
                for qubit in targets
            }
        )

    options = {
        "nshots": params.nshots,
        "relaxation_time": params.relaxation_time,
        "acquisition_type": AcquisitionType.DISCRIMINATION,
        "averaging_mode": AveragingMode.SINGLESHOT,
    }

    # execute the pulse sequence
    if params.unrolling:
        results = platform.execute(sequences, **options)
        for beta, ro_pulses in zip(beta_param_range, all_ro_pulses):
            for qubit in targets:
                result = results[ro_pulses[qubit].id]
                prob = probability(result, state=0)
                # store the results
                data.register_qubit(
                    DragTuningType,
                    (qubit),
                    dict(
                        prob=np.array([prob]),
                        error=np.array([np.sqrt(prob * (1 - prob) / params.nshots)]),
                        beta=np.array([beta]),
                    ),
                )
    else:
        for i, sequence in enumerate(sequences):
            result = platform.execute([sequence], **options)
            for qubit in targets:
                ro_pulse = list(sequence.channel(platform.qubits[qubit].acquisition))[
                    -1
                ]
                prob = probability(result[ro_pulse.id], state=0)
                # store the results
                data.register_qubit(
                    DragTuningType,
                    (qubit),
                    dict(
                        prob=np.array([prob]),
                        error=np.array([np.sqrt(prob * (1 - prob) / params.nshots)]),
                        beta=np.array([beta_param_range[i]]),
                    ),
                )

    return data


def drag_fit(x, offset, amplitude, period, phase):
    return offset + amplitude * np.cos(2 * np.pi * x / period + phase)


def _fit(data: DragTuningData) -> DragTuningResults:
    qubits = data.qubits
    betas_optimal = {}
    fitted_parameters = {}
    chi2 = {}

    for qubit in qubits:
        qubit_data = data[qubit]

        # normalize prob
        prob = qubit_data.prob
        prob_min = np.min(prob)
        prob_max = np.max(prob)
        normalized_prob = (prob - prob_min) / (prob_max - prob_min)

        # normalize beta
        beta_params = qubit_data.beta
        beta_min = np.min(beta_params)
        beta_max = np.max(beta_params)
        normalized_beta = (beta_params - beta_min) / (beta_max - beta_min)

        # Guessing period using fourier transform
        period = fallback_period(guess_period(normalized_beta, normalized_prob))
        pguess = [0.5, 0.5, period, 0]
        try:
            popt, _ = curve_fit(
                drag_fit,
                normalized_beta,
                normalized_prob,
                p0=pguess,
                maxfev=100000,
                bounds=(
                    [0, 0, 0, -np.pi],
                    [1, 1, np.inf, np.pi],
                ),
                sigma=qubit_data.error,
            )
            translated_popt = [
                popt[0] * (prob_max - prob_min) + prob_min,
                popt[1] * (prob_max - prob_min),
                popt[2] * (beta_max - beta_min),
                popt[3] - 2 * np.pi * beta_min / popt[2] / (beta_max - beta_min),
            ]
            fitted_parameters[qubit] = translated_popt
            predicted_prob = drag_fit(beta_params, *translated_popt)
            betas_optimal[qubit] = beta_params[np.argmax(predicted_prob)]
            chi2[qubit] = (
                chi2_reduced(
                    prob,
                    predicted_prob,
                    qubit_data.error,
                ),
                np.sqrt(2 / len(prob)),
            )
        except Exception as e:
            log.warning(f"drag_tuning_fit failed for qubit {qubit} due to {e}.")
    return DragTuningResults(betas_optimal, fitted_parameters, chi2=chi2)


def _plot(data: DragTuningData, target: QubitId, fit: DragTuningResults):
    """Plotting function for DragTuning."""

    figures = []
    fitting_report = ""

    qubit_data = data[target]
    betas = qubit_data.beta
    fig = go.Figure(
        [
            go.Scatter(
                x=qubit_data.beta,
                y=qubit_data.prob,
                opacity=1,
                mode="lines",
                name="Probability",
                showlegend=True,
                legendgroup="Probability",
            ),
            go.Scatter(
                x=np.concatenate((betas, betas[::-1])),
                y=np.concatenate(
                    (
                        qubit_data.prob + qubit_data.error,
                        (qubit_data.prob - qubit_data.error)[::-1],
                    )
                ),
                fill="toself",
                fillcolor=COLORBAND,
                line=dict(color=COLORBAND_LINE),
                showlegend=True,
                name="Errors",
            ),
        ]
    )

    # add fitting traces
    if fit is not None:
        beta_range = np.linspace(
            min(betas),
            max(betas),
            20,
        )

        fig.add_trace(
            go.Scatter(
                x=beta_range,
                y=drag_fit(beta_range, *fit.fitted_parameters[target]),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            ),
        )
        fitting_report = table_html(
            table_dict(
                target,
                ["Optimal Beta Param", "Chi2 reduced"],
                [(np.round(fit.betas[target], 4), 0), fit.chi2[target]],
                display_error=True,
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Beta parameter",
        yaxis_title="Ground State Probability",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(results: DragTuningResults, platform: CalibrationPlatform, target: QubitId):
    update.drag_pulse_beta(
        results.betas[target],
        platform,
        target,
    )


drag_tuning = Routine(_acquisition, _fit, _plot, _update)
"""DragTuning Routine object."""



from dataclasses import dataclass, field
from typing import Union

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
from qibolab import (
    AcquisitionType,
    AveragingMode,
    Delay,
    Parameter,
    PulseSequence,
    Sweeper,
)

from qibocal.auto.operation import Data, Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform
from qibocal.result import magnitude, phase

from ... import update
from ..utils import readout_frequency, table_dict, table_html
from . import utils


@dataclass
class T1SignalParameters(Parameters):
    """T1 runcard inputs."""

    delay_before_readout_start: int
    """Initial delay before readout [ns]."""
    delay_before_readout_end: int
    """Final delay before readout [ns]."""
    delay_before_readout_step: int
    """Step delay before readout [ns]."""
    single_shot: bool = False
    """If ``True`` save single shot signal data."""


@dataclass
class T1SignalResults(Results):
    """T1 Signal outputs."""

    t1: dict[QubitId, Union[float, list[float]]]
    """T1 for each qubit."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""
    pcov: dict[QubitId, list[float]]
    """Approximate covariance of fitted parameters."""


@dataclass
class T1SignalData(Data):
    """T1 acquisition outputs."""

    data: dict[QubitId, npt.NDArray] = field(default_factory=dict)
    """Raw data acquired."""

    @property
    def average(self):
        if len(next(iter(self.data.values())).shape) > 1:
            return utils.average_single_shots(self.__class__, self.data)
        return self


def t1_sequence(platform: CalibrationPlatform, targets: list[QubitId]):
    """Create sequence for T1 experiment with a given optional delay."""
    sequence = PulseSequence()
    ro_pulses, delays = {}, {}
    for q in targets:
        natives = platform.natives.single_qubit[q]
        qd_channel, qd_pulse = natives.RX()[0]
        ro_channel, ro_pulse = natives.MZ()[0]

        ro_pulses[q] = ro_pulse
        delays[q] = Delay(duration=0)

        sequence.append((qd_channel, qd_pulse))
        sequence.append((ro_channel, Delay(duration=qd_pulse.duration)))
        sequence.append((ro_channel, delays[q]))
        sequence.append((ro_channel, ro_pulse))

    return sequence, ro_pulses, delays


def _acquisition(
    params: T1SignalParameters, platform: CalibrationPlatform, targets: list[QubitId]
) -> T1SignalData:
    """Data acquisition for T1 experiment.

    In this protocol the y axis is the magnitude of signal in the IQ plane."""

    sequence, ro_pulses, delays = t1_sequence(platform, targets)

    ro_wait_range = np.arange(
        params.delay_before_readout_start,
        params.delay_before_readout_end,
        params.delay_before_readout_step,
    )

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=ro_wait_range,
        pulses=[delays[q] for q in targets],
    )

    results = platform.execute(
        [sequence],
        [[sweeper]],
        updates=[
            {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
            for q in targets
        ],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=(
            AveragingMode.SINGLESHOT if params.single_shot else AveragingMode.CYCLIC
        ),
    )

    data = T1SignalData()

    for q in targets:
        result = results[ro_pulses[q].id]
        signal = magnitude(result)
        if params.single_shot:
            _waits = np.array(len(signal) * [ro_wait_range])
        else:
            _waits = ro_wait_range
        data.register_qubit(
            utils.CoherenceType,
            (q),
            dict(wait=_waits, signal=signal, phase=phase(result)),
        )

    return data


def _fit(data: T1SignalData) -> T1SignalResults:
    """
    Fitting routine for T1 experiment. The used model is

        .. math::

            y = p_0-p_1 e^{-x p_2}.
    """
    data = data.average
    t1s, fitted_parameters, pcovs = utils.exponential_fit(data)

    return T1SignalResults(t1s, fitted_parameters, pcovs)


def _plot(data: T1SignalData, target: QubitId, fit: T1SignalResults = None):
    """Plotting function for T1 experiment."""
    data = data.average

    figures = []
    fig = go.Figure()

    fitting_report = None
    qubit_data = data[target]
    waits = qubit_data.wait

    fig.add_trace(
        go.Scatter(
            x=waits,
            y=qubit_data.signal,
            opacity=1,
            name="Signal",
            showlegend=True,
            legendgroup="Signal",
        )
    )

    if fit is not None:
        waitrange = np.linspace(
            min(waits),
            max(waits),
            2 * len(qubit_data),
        )

        params = fit.fitted_parameters[target]
        fig.add_trace(
            go.Scatter(
                x=waitrange,
                y=utils.exp_decay(waitrange, *params),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                target, ["T1 [ns]"], [np.round(fit.t1[target])], display_error=True
            )
        )

    # last part
    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Signal [a.u.]",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(results: T1SignalResults, platform: CalibrationPlatform, target: QubitId):
    update.t1(results.t1[target], platform, target)


t1_signal = Routine(_acquisition, _fit, _plot, _update)
"""T1 Signal Routine object."""



from dataclasses import dataclass
from typing import Union

import numpy as np
import plotly.graph_objects as go
from qibolab import AcquisitionType, AveragingMode, Parameter, Sweeper

from qibocal import update
from qibocal.auto.operation import Parameters, QubitId, Results, Routine
from qibocal.calibration import CalibrationPlatform

from ...result import magnitude, phase
from ..ramsey.utils import ramsey_sequence
from ..utils import readout_frequency, table_dict, table_html
from . import t1_signal, utils


@dataclass
class T2SignalParameters(Parameters):
    """T2Signal runcard inputs."""

    delay_between_pulses_start: int
    """Initial delay between RX(pi/2) pulses in ns."""
    delay_between_pulses_end: int
    """Final delay between RX(pi/2) pulses in ns."""
    delay_between_pulses_step: int
    """Step delay between RX(pi/2) pulses in ns."""
    single_shot: bool = False
    """If ``True`` save single shot signal data."""


@dataclass
class T2SignalResults(Results):
    """T2Signal outputs."""

    t2: dict[QubitId, Union[float, list[float]]]
    """T2 for each qubit [ns]."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""
    pcov: dict[QubitId, list[float]]
    """Approximate covariance of fitted parameters."""


class T2SignalData(t1_signal.T1SignalData):
    """T2Signal acquisition outputs."""

    t2: dict[QubitId, float]
    """T2 for each qubit [ns]."""
    fitted_parameters: dict[QubitId, dict[str, float]]
    """Raw fitting output."""


def _acquisition(
    params: T2SignalParameters,
    platform: CalibrationPlatform,
    targets: list[QubitId],
) -> T2SignalData:
    """Data acquisition for T2 experiment.

    In this protocol the y axis is the magnitude of signal in the IQ plane.

    """

    waits = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    sequence, delays = ramsey_sequence(platform, targets)

    data = T2SignalData()

    sweeper = Sweeper(
        parameter=Parameter.duration,
        values=waits,
        pulses=delays,
    )

    results = platform.execute(
        [sequence],
        [[sweeper]],
        updates=[
            {platform.qubits[q].probe: {"frequency": readout_frequency(q, platform)}}
            for q in targets
        ],
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
        acquisition_type=AcquisitionType.INTEGRATION,
        averaging_mode=(
            AveragingMode.SINGLESHOT if params.single_shot else AveragingMode.CYCLIC
        ),
    )

    for q in targets:
        ro_pulse = list(sequence.channel(platform.qubits[q].acquisition))[-1]
        result = results[ro_pulse.id]
        signal = magnitude(result)
        if params.single_shot:
            _waits = np.array(len(signal) * [waits])
        else:
            _waits = waits
        data.register_qubit(
            utils.CoherenceType,
            (q),
            dict(wait=_waits, signal=signal, phase=phase(result)),
        )
    return data


def _fit(data: T2SignalData) -> T2SignalResults:
    """The used model is

    .. math::

        y = p_0 - p_1 e^{-x p_2}.
    """
    data = data.average

    t2s, fitted_parameters, pcovs = utils.exponential_fit(data)
    return T2SignalResults(t2s, fitted_parameters, pcovs)


def _plot(data: T2SignalData, target: QubitId, fit: T2SignalResults = None):
    """Plotting function for Ramsey Experiment."""
    data = data.average

    figures = []
    fig = go.Figure()
    fitting_report = None

    qubit_data = data[target]

    fig.add_trace(
        go.Scatter(
            x=qubit_data.wait,
            y=qubit_data.signal,
            opacity=1,
            name="Signal",
            showlegend=True,
            legendgroup="Signal",
        )
    )

    if fit is not None:
        # add fitting trace
        waitrange = np.linspace(
            min(qubit_data.wait),
            max(qubit_data.wait),
            2 * len(qubit_data),
        )

        params = fit.fitted_parameters[target]
        fig.add_trace(
            go.Scatter(
                x=waitrange,
                y=utils.exp_decay(
                    waitrange,
                    *params,
                ),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
            )
        )
        fitting_report = table_html(
            table_dict(
                target, ["T2 [ns]"], [np.round(fit.t2[target])], display_error=True
            )
        )

    fig.update_layout(
        showlegend=True,
        xaxis_title="Time [ns]",
        yaxis_title="Signal [a.u.]",
    )

    figures.append(fig)

    return figures, fitting_report


def _update(results: T2SignalResults, platform: CalibrationPlatform, target: QubitId):
    update.t2(results.t2[target], platform, target)


t2_signal = Routine(_acquisition, _fit, _plot, _update)
"""T2Signal Routine object."""



from .allxy.allxy import allxy
from .allxy.allxy_resonator_depletion_tuning import allxy_resonator_depletion_tuning
from .classification import single_shot_classification
from .coherence.cpmg import cpmg
from .coherence.spin_echo import spin_echo
from .coherence.spin_echo_signal import spin_echo_signal
from .coherence.t1 import t1
from .coherence.t1_signal import t1_signal
from .coherence.t2 import t2
from .coherence.t2_signal import t2_signal
from .coherence.zeno import zeno
from .dispersive_shift import dispersive_shift
from .dispersive_shift_qutrit import dispersive_shift_qutrit
from .drag import drag_tuning
from .drag_simple import drag_simple
from .flipping import flipping
from .flux_amplitude_frequency import flux_amplitude_frequency
from .flux_dependence.qubit_crosstalk import qubit_crosstalk
from .flux_dependence.qubit_flux_dependence import qubit_flux
from .flux_dependence.resonator_flux_dependence import resonator_flux
from .flux_gate import flux_gate
from .qubit_power_spectroscopy import qubit_power_spectroscopy
from .qubit_spectroscopy import qubit_spectroscopy
from .qubit_spectroscopy_ef import qubit_spectroscopy_ef
from .qutrit_classification import qutrit_classification
from .rabi.amplitude import rabi_amplitude
from .rabi.amplitude_frequency import rabi_amplitude_frequency
from .rabi.amplitude_frequency_signal import rabi_amplitude_frequency_signal
from .rabi.amplitude_signal import rabi_amplitude_signal
from .rabi.ef import rabi_amplitude_ef
from .rabi.length import rabi_length
from .rabi.length_frequency import rabi_length_frequency
from .rabi.length_frequency_signal import rabi_length_frequency_signal
from .rabi.length_signal import rabi_length_signal
from .ramsey.ramsey import ramsey
from .ramsey.ramsey_signal import ramsey_signal
from .ramsey.ramsey_zz import ramsey_zz
from .randomized_benchmarking.filtered_rb import filtered_rb
from .randomized_benchmarking.standard_rb import standard_rb
from .randomized_benchmarking.standard_rb_2q import standard_rb_2q
from .randomized_benchmarking.standard_rb_2q_inter import standard_rb_2q_inter
from .readout_characterization import readout_characterization
from .readout_mitigation_matrix import readout_mitigation_matrix
from .readout_optimization.resonator_amplitude import resonator_amplitude
from .resonator_punchout import resonator_punchout
from .resonator_spectroscopy import resonator_spectroscopy
from .signal_experiments.calibrate_state_discrimination import (
    calibrate_state_discrimination,
)
from .signal_experiments.time_of_flight_readout import time_of_flight_readout
from .state_tomography import state_tomography
from .two_qubit_interaction import (
    chevron,
    chevron_signal,
    correct_virtual_z_phases,
    cryoscope,
    optimize_two_qubit_gate,
)
from .two_qubit_state_tomography import two_qubit_state_tomography

__all__ = [
    "allxy",
    "single_shot_classification",
    "spin_echo",
    "spin_echo_signal",
    "t1",
    "t1_signal",
    "t2",
    "t2_signal",
    "zeno",
    "dispersive_shift",
    "dispersive_shift_qutrit",
    "drag_tuning",
    "flipping",
    "qubit_crosstalk",
    "qubit_flux",
    "resonator_flux",
    "qubit_spectroscopy",
    "qubit_spectroscopy_ef",
    "qutrit_classification",
    "rabi_amplitude",
    "rabi_amplitude_signal",
    "rabi_length",
    "rabi_amplitude_ef",
    "rabi_length_signal",
    "ramsey",
    "ramsey_signal",
    "filtered_rb",
    "standard_rb",
    "readout_characterization",
    "readout_mitigation_matrix",
    "resonator_amplitude",
    "resonator_punchout",
    "resonator_spectroscopy",
    "calibrate_state_discrimination",
    "time_of_flight_readout",
    "chevron",
    "chevron_signal",
    "correct_virtual_z_phases",
    "state_tomography",
    "allxy_resonator_depletion_tuning",
    "two_qubit_state_tomography",
    "qubit_power_spectroscopy",
    "rabi_amplitude_frequency",
    "rabi_amplitude_frequency_signal",
    "rabi_length_frequency",
    "rabi_length_frequency_signal",
    "standard_rb_2q",
    "standard_rb_2q_inter",
    "optimize_two_qubit_gate",
    "cryoscope",
    "ramsey_zz",
    "flux_gate",
    "flux_amplitude_frequency",
    "cpmg",
    "drag_simple",
]
